/***********************************************************************

	$Id: datum.c,v 1.4 2009/09/02 10:31:08 warme Exp $

	File:	datum.c
	Rev:	a-1
	Date:	12/17/2005

	Copyright (c) 2005, 2009 by David M. Warme.  All rights
	reserved.

************************************************************************

	Data structures to represent numeric input values.

************************************************************************

	Modification Log:

	a-1:	12/17/2005	warme
		: Created.

************************************************************************/

#include "datum.h"
#include "cbuf.h"
#include "logic.h"
#include "memory.h"
#include "mystring.h"

#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

extern char *	me;

/*
 * Read a single datum from the current line of the given input file.
 *
 * Return values:
 *
 *	-1	End Of File was encountered before a datum was read.
 *	0	End Of Line was encountered before a datum was read.
 *	+1	A datum was successfully read.
 *
 * If a comment character is encountered, strip to the end of line and
 * return 0.
 */

	int
read_datum (

FILE *		fp,		/* IN - file to read from */
struct Datum *	datum		/* OUT - datum that was read */
)
{
int		c;
char *		p;
char *		p2;
char *		p3;
double		negative;
struct CBuf	buf;

	init_cbuf (&buf);

	for (;;) {
		c = fgetc (fp);
		if (c < 0) return (-1);

		if ((c EQ ' ') OR (c EQ '\t') OR (c EQ '\f')) {
			if (buf.count <= 0) continue;
			ungetc (c, fp);
			break;
		}

		if (c EQ '\n') {
			if (buf.count <= 0) {
				return (0);
			}
			ungetc (c, fp);
			break;
		}

		if ((c EQ '@') OR (c EQ '#') OR (c EQ '%')) {
			if (buf.count > 0) {
				/* Comment ends existing field. */
				ungetc (c, fp);
				break;
			}

			/* Strip comment and indicate EOL. */
			do {
				c = fgetc (fp);
				if (c < 0) return (-1);
			} while (c NE '\n');
			return (0);
		}

		if ((c EQ '+') OR
		    (c EQ '-') OR
		    (c EQ '.') OR
		    ((c >= '0') AND (c <= '9')) OR
		    ((c >= 'A') AND (c <= 'Z')) OR
		    ((c >= 'a') AND (c <= 'z'))) {
			put_cbuf (&buf, c);
		}
		else {
			fprintf (stderr,
				 "%s: Invalid char '%c' in data field\n",
				 me, c);
			exit (1);
		}
	}

	put_cbuf (&buf, '\0');

	if (buf.next NE NULL) {
		/* This data field is MUCH too long to handle! */
		fprintf (stderr, "%s: Data field too wide!\n", me);
		exit (1);
	}

	/* Handle optional sign. */
	p = &buf.buf [0];
	negative = 0;
	if (*p EQ '+') {
		++p;
	}
	else if (*p EQ '-') {
		negative = 1;
		++p;
	}

	/* Identify the various special cases. */
	if (mystrcasecmp (p, "inf") EQ 0) {
		if (NOT negative) {
			datum -> type	= DATUM_TYPE_PINF;
		}
		else {
			datum -> type	= DATUM_TYPE_MINF;
		}
		datum -> value = 0.0;
		return (1);
	}
	if (mystrcasecmp (p, "nan") EQ 0) {
		datum -> type	= DATUM_TYPE_NAN;
		datum -> value	= 0.0;
		return (1);
	}
	c = *p;
	if (((c >= '0') AND (c <= '9')) OR (c EQ '.')) {
		datum -> type	= DATUM_TYPE_VALUE;
		if (negative) {
			datum -> value	= - strtod (p, &p2);
		}
		else {
			datum -> value	= strtod (p, &p2);
		}
		p3 = &buf.buf [buf.count - 1];
		if (p2 NE p3) {
			fprintf (stderr,
				 "%s: Invalid numeric data `%s'\n",
				 me, buf.buf);
			exit (1);
		}
	}
	else {
		fprintf (stderr,
			 "%s: Invalid numeric data `%s'\n",
			 me, buf.buf);
		exit (1);
	}

	return (1);
}

/*
 * Append a single datum to the given field buffer.
 */

	void
append_datum (

struct DBuf **		hookp,		/* IN/OUT - buffer pointer hook */
struct Datum *		datum		/* IN - datum to append */
)
{
struct DBuf *	bufp;
struct DBuf *	last;
struct DBuf *	newp;

	bufp = *hookp;
	if (bufp EQ NULL) {
		bufp = NEW (struct DBuf);
		bufp -> count	= 0;
		bufp -> next	= NULL;
		bufp -> last	= NULL;
		*hookp = bufp;
	}

	last = bufp -> last;
	if (last EQ NULL) {
		last = bufp;
	}

	if (last -> count >= MAX_DBUF) {
		/* Need to append another buffer onto the end. */
		newp = NEW (struct DBuf);
		newp -> count	= 0;
		newp -> next	= NULL;
		newp -> last	= NULL;

		if (last EQ bufp) {
			bufp -> next = newp;
			bufp -> last = newp;
		}
		else {
			last -> next = newp;
			bufp -> last = newp;
		}
		last = newp;
	}
	last -> buf [last -> count++] = *datum;
}

/*
 * Create a linear array of Datum object from the given DBuf.
 */

	void
linearize_dbuf (

struct DBuf *	bufp,		/* IN - DBuf to linearize */
struct Datum **	array,		/* OUT - linear array of Datum objects */
int *		count		/* OUT - number of elements in array */
)
{
int		i, k, n;
struct DBuf *	p;
struct Datum *	dp;

	k = 0;
	for (p = bufp; p NE NULL; p = p -> next) {
		k += p -> count;
	}

	dp = NULL;
	if (k > 0) {
		dp = NEWA (k, struct Datum);
		memset (dp, 0, k * sizeof (dp [0]));
	}
	*array = dp;
	*count = k;

	for (p = bufp; p NE NULL; p = p -> next) {
		n = p -> count;
		for (i = 0; i < n; i++) {
			*dp++ = p -> buf [i];
		}
	}
}

/*
 * Free a chain of DBuf's.
 */

	void
free_dbuf (

struct DBuf *	bufp
)
{
struct DBuf *	tmp;

	while (bufp NE NULL) {
		tmp = bufp -> next;
		free (bufp);
		bufp = tmp;
	}
}
