/***********************************************************************

	$Id: eeg.c,v 1.5 2009/09/02 10:49:42 warme Exp $

	File:	eeg.c
	Rev:	a-4
	Date:	09/02/2009

	Copyright (c) 2005, 2009 by David M. Warme.  All rights
	reserved.

************************************************************************

	A "brainwave" style strip-chart recorder.  Converts a
	sequence of N lines, each containing K IEEE floating point
	numbers into a "K-channel" postscript strip-chart plot,
	possibly spread across several printed pages.

************************************************************************

	Modification Log:

	a-1:	12/17/2005	warme
		: Created.
	a-2:	05/29/2006	warme
		: Added new configuration file.
		: Now using simple @EEG_TYPE@ "type" "version"
		:  marker instead of previous complicated
		:  S-expression stuff (which I eliminated).
	a-3:	06/30/2007	warme
		: Final record wasn't being displayed.
	a-4:	09/02/2009	warme
		: Added "linear" channel option.
		: Added -v flag.
		: Added autoconf/config.h stuff.
		: Fix generated postscript to properly describe
		:  the paper size used.

************************************************************************/

#define _GNU_SOURCE

#include <errno.h>
#include <limits.h>
#include <math.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#include "cbuf.h"
#include "config.h"
#include "datum.h"
#include "logic.h"
#include "memory.h"
#include "mystring.h"
#include "name.h"
#include "paper.h"
#include "pnodes.h"
#include "ps_prolog.h"


/*
 * Global Routines
 */

int			main (int, char **);

char *			me;

extern FILE *		yyin;


/*
 * Local Types
 */

#define	MARGIN		36.0		/* Page margin. */

#define MAX_CHANNELS	256

enum PageMode {
	PAGEMODE_LANDSCAPE,
	PAGEMODE_PORTRAIT
};


struct ChannelDef {	/* Info about a single data field */
	double		mag_factor;	/* Magnification factor */
	double		linear_min;	/* Minimum of linear scale */
	double		linear_max;	/* Maximum of linear scale */
	int		force_linear;	/* Force scale to linear? */
	int		fldno;		/* Index of field to plot */
	const char *	title;		/* Title of field */
	int		scatter;	/* Scatter plot this data field? */
};

struct BScale {
	int		logscale;	/* TRUE ==> log, FALSE ==> linear */
	int		minpow;		/* min power of 10 */
	int		maxpow;		/* max power of 10 */
	int		powtics;	/* Y tic every 10**powtics */
	double		minval;		/* min linear scale value */
	double		maxval;		/* max linear scale value */
	double		ytics;		/* Y tic interval */
};

struct StripInfo {
	int		field;		/* Which field is displayed */
	int		scatter;	/* If true, do scatter plot instead */
					/* plotting lines between data points */
	double		mag_factor;	/* Magnification factor */
	double		top_y;		/* Vertical position of top of */
					/* this strip on the page */
	double		bot_y;		/* Vertical position of bottom of */
					/* this strip on the page */
	double		scale_ymax;	/* Vertical pos of top of scale labels */
	double		scale_ymin;	/* Vertical pos of bottom of scale labels */
	double		linear_min;	/* Minimum, if forcing linear scale */
	double		linear_max;	/* Maximum, if forcing linear scale */
	int		force_linear;	/* Force linear scale, if TRUE */
	const char *	title;		/* Title string for strip */
	struct BScale	pband;		/* Scaling info for positive band */
	struct BScale	mband;		/* Scaling info for negative band */
	/* Band0 ==> +inf, Band1 ==> plus, Band2 ==> zero, Band3 ==> minus, */
	/* Band4 ==> -inf, Band5 ==> nan. */
	int		top_band;	/* Top band to display */
	int		bot_band;	/* Bottom band to display */
	double		pinf_pos;	/* Where to draw +inf line */
	double		pinf_label_top;	/* Top of +inf label box */
	double		pinf_label_bot;	/* Bottom of +inf label box */
	double		plus_top_pos;	/* Top of positive band */
	double		plus_bot_pos;	/* Bottom of positive band */
	double		plus_label_top;	/* Top of positive band labels box */
	double		plus_label_bot;	/* Bottom of positive band labels box */
	double		zero_pos;	/* Where to draw zero line */
	double		zero_label_top;	/* Top of zero label box */
	double		zero_label_bot;	/* Bottom of zero label box */
	double		minus_top_pos;	/* Top of negative band */
	double		minus_bot_pos;	/* Bottom of negative band */
	double		minus_label_top; /* Top of negative band labels box */
	double		minus_label_bot; /* Bottom of negative band labels box */
	double		minf_pos;	/* Where to draw -inf line */
	double		minf_label_top;	/* Top of -inf label box */
	double		minf_label_bot;	/* Bottom of -inf label box */
	double		nan_pos;	/* Where to draw nan line */
	double		nan_label_top;	/* Top of nan label box */
	double		nan_label_bot;	/* Bottom of nan label box */
};

struct Layout {
	double	xmin;			/* Right end of left margin */
	double	xmax;			/* Left end of right margin */

	double	top_title_top_y;	/* Top edge of top title box */
	double	top_title_bot_y;	/* Bottom edge of top title box */
	double	bottom_title_top_y;	/* Top edge of bottom title box */
	double	bottom_title_bot_y;	/* Bottom edge of bottom title box */

	double	bottom_scale_ymax;	/* Top edge of bottom xscale box */
	double	bottom_scale_ymin;	/* Bottom edge of bottom xscale box */

	double	chan_label_xmin;	/* Left  end of channel label box */
	double	chan_label_xmax;	/* Right end of channel label box */
	double	left_scale_xmin;	/* Left  end of left scale values */
	double	left_scale_xmax;	/* Right end of left scale values */
	double	left_tick_xmin;		/* Left  end of left tick marks */
	double	left_tick_xmax;		/* Right end of left tick marks */
	double	plot_area_xmin;		/* Left  end of plotting area */
	double	plot_area_xmax;		/* Right end of plotting area */
	double	data_xmin;		/* Left  end of actual plotted data */
	double	data_xmax;		/* Right end of actual plotted data */
	double	right_tick_xmin;	/* Left  end of right tick marks */
	double	right_tick_xmax;	/* Right end of right tick marks */
	double	right_scale_xmin;	/* Left  end of right scale values */
	double	right_scale_xmax;	/* Right end of right scale values */
};


/*
 * Local Routines
 */

static void		box (double xmin, double xmax, double ymin, double ymax);
static void		choose_x_axis ();
static void		choose_y_axes ();
static void		choose_y_axis (struct StripInfo * sip);
static char *		cvint (char * ap, int * chan, int min);
static double		datum_y_value (struct StripInfo *	sip,
				       const struct Datum *	dp);
static void		debug_print_parse_tree (void);
static void		decode_channel_order_arg (char * ap);
static void		decode_params (int argc, char ** argv);
static void		define_strips (void);
static void		emit_postscript_string (const char *);
static void		emit_prolog_section (void);
static void		emit_comments_section (void);
static void		emit_plot_page (void);
static void		emit_setup_section (void);
static void		empty_numeric_band (struct BScale * bscale);
static void		isolate_selected_data (void);
static void		parse_config_file (void);
static void		partition_page_into_horizontal_strips (void);
static void		partition_page_into_vertical_strips (void);
static void		plot_strip_data (struct StripInfo * sip);
static void		plot_linear_y_scale (struct StripInfo * sip,
					     struct BScale *	bscale,
					     int		sign,
					     double		ymax,
					     double		ymin,
					     double		label_ymax,
					     double		label_ymin);
static void		plot_log_y_scale (struct StripInfo * sip,
					  struct BScale *	bscale,
					  int		 	sign,
					  double		ymax,
					  double		ymin,
					  double		label_ymax,
					  double		label_ymin);
static void		plot_x_ticks_and_scale (void);
static void		plot_y_scale (struct StripInfo * sip,
				      struct BScale *	 bscale,
				      int		 sign,
				      double		 ymax,
				      double		 ymin,
				      double		 label_ymax,
				      double		 label_ymin);
static void		process_option_list (struct PlotOptionPnode * options);
static void		process_selected_plot (const char * template_name);
static void		process_type_and_version (struct Name *	tp,
						  struct Name *	vp);
static void		put_data_into_arrays (void);
static void		read_data (FILE * fp);
static void		read_marker (FILE * fp);
static struct Name *	read_quoted_string (FILE * fp);
static void		scale_numeric_band (double		minmag,
					    double		maxmag,
					    struct StripInfo *	sip,
					    int			pband);
static double		scale_numeric_value (double		val,
					     struct BScale *	bsp);
static void		usage (void);


/*
 * Local Variables
 */

static int		config_nfields;
static struct SourceDefPnode *	data_type;
static struct VersionDefPnode *	data_version;
static const char *	flag_config_file_path;
static struct Datum *	field_data [MAX_CHANNELS];
static struct DBuf *	field_dbufs [MAX_CHANNELS];
static struct Name *	field_name [MAX_CHANNELS];
static const char *	flag_bottom_title;
static int		flag_channel_order [MAX_CHANNELS];
static int		flag_channel_order_count;
static int		flag_mag_count;
static double		flag_mag_factor [MAX_CHANNELS];
static int		flag_mag_strip_num [MAX_CHANNELS];
static int		flag_numpages;
static enum PageMode	flag_pagemode;
static int		flag_pagemode_given;
static int		flag_scatter_plot [MAX_CHANNELS];
static const struct PaperSize * flag_papersize;
static char *		flag_template;
static const char *	flag_top_title;
static int		flag_version;
static struct Layout	layout;
static int		nfields;
static int		nrecords;
static int		nstrips;
static const char *	option_bottom_title;
static struct ChannelDef option_chan_defs [MAX_CHANNELS];
static int		option_chan_def_count;
static enum PageMode	option_pagemode;
static int		option_pagemode_given;
static const char *	option_top_title;
static double		scale_xmax;
static double		scale_xmin;
static struct StripInfo	strip_info [MAX_CHANNELS];
static int		xmax;
static int		xmin;
static double		scale_xtics;

/*
 * The main routine for the "eeg" program.
 */

	int
main (

int		argc,
char **		argv
)
{
FILE *		fp;

	me = argv [0];

	decode_params (argc, argv);

	if (flag_version) {
		printf ("eeg version %s\n", PACKAGE_VERSION_STRING);
		exit (0);
	}

	parse_config_file ();

	fp = stdin;

	read_marker (fp);

	read_data (fp);

	put_data_into_arrays ();

	process_selected_plot (flag_template);

	isolate_selected_data ();

	define_strips ();

	partition_page_into_horizontal_strips ();

	partition_page_into_vertical_strips ();

	choose_x_axis ();

	choose_y_axes ();

	emit_comments_section ();
	emit_prolog_section ();
	emit_setup_section ();

	emit_plot_page ();

	exit (0);
}

/*
 * This routine decodes the various command-line arguments.
 */

	static
	void
decode_params (

int		argc,
char **		argv
)
{
int		strip;
int		x, y;
char *		ap;
char		c;
double		factor;

	flag_bottom_title		= NULL;
	flag_channel_order_count	= 0;
	flag_mag_count			= 0;
	flag_numpages			= 1;
	flag_pagemode			= PAGEMODE_PORTRAIT;
	flag_pagemode_given		= FALSE;
	flag_top_title			= NULL;
	flag_version			= FALSE;

	/* Default to letter size paper. */
	flag_papersize = get_default_paper_size ();

	xmin = -1;
	xmax = -1;

	--argc;
	me = *argv++;

	while (argc > 0) {
		ap = *argv++;
		if (*ap NE '-') {
			usage ();
		}
		++ap;
		while ((c = *ap++) NE '\0') {
			switch (c) {
			case 'b':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					--argc;
				}
				flag_bottom_title = ap;
				ap = NULL;
				break;

			case 'c':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					--argc;
				}
				decode_channel_order_arg (ap);
				ap = NULL;
				break;

			case 'f':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					--argc;
				}
				flag_config_file_path = ap;
				ap = NULL;
				break;

			case 'l':
				flag_pagemode		= PAGEMODE_LANDSCAPE;
				flag_pagemode_given	= TRUE;
				break;

			case 'm':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					--argc;
				}
				ap = cvint (ap, &strip, 1);
				if (*ap NE ',') usage ();
				++ap;
				factor = atof (ap);
				if (flag_mag_count >= MAX_CHANNELS) usage ();
				flag_mag_strip_num [flag_mag_count] = strip;
				flag_mag_factor [flag_mag_count] = factor;
				++flag_mag_count;
				ap = NULL;
				break;

			case 'P':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					--argc;
				}
				flag_numpages = atoi (ap);
				if (flag_numpages < 1) usage ();
				ap = NULL;
				break;

			case 'p':
				flag_pagemode		= PAGEMODE_PORTRAIT;
				flag_pagemode_given	= TRUE;
				break;

			case 'S':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					--argc;
				}
				strip = atoi (ap);
				if ((strip < 1) OR (strip > MAX_CHANNELS)) {
					fprintf (stderr,
						 "%s: Strip %d is out of"
						 " range [1,%d] (arg -S %d)\n",
						 strip, MAX_CHANNELS);
					exit (1);
				}
				flag_scatter_plot [strip - 1] = TRUE;
				ap = NULL;
				break;

			case 's':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					--argc;
				}
				flag_papersize = get_paper_size (ap);
				ap = NULL;
				break;

			case 't':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					--argc;
				}
				flag_top_title = ap;
				ap = NULL;
				break;

			case 'T':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					--argc;
				}
				flag_template = ap;
				ap = NULL;
				break;

			case 'v':
				flag_version = TRUE;
				break;

			case 'x':
				if (*ap EQ '\0') {
					if (argc <= 0) {
						usage ();
					}
					ap = *argv++;
					--argc;
				}
				ap = cvint (ap, &xmin, 0);
				if ((*ap NE ',') AND (*ap NE ':')) usage ();
				++ap;
				xmax = atoi (ap);
				if (xmin >= xmax) {
					fprintf (stderr,
						 "%s: xmin >= xmax (%d >= %d)\n",
						 me, xmin, xmax);
					exit (1);
				}
				ap = NULL;
				break;

			default:
				usage ();
				break;
			}
			if (ap EQ NULL) break;
		}
		--argc;
	}

	if ((flag_template NE NULL) AND (flag_channel_order_count > 0)) {
		/* Cannot mix channel specs from the command line	*/
		/* and from the configuration file!			*/
		fprintf (stderr,
			 "%s: The -T and -c flags cannot be used together.\n",
			 me);
		exit (1);
	}
}

/*
 * Parse the channel order argument string.  This is a comma-separated
 * sequence of items of the form: N, N-M, N-, or -M, where N and M are
 * positive decimal integer constants, and N <= N.
 *
 * Because we have not yet read any of the input data (at the time we
 * process command line arguments), we do not know how many channels of
 * data actually appear in the input data.  Therefore it is not possible
 * to error check the N/M values, nor can we really know what "N-"
 * really means.  When we see "N-", we simply store channel N-1 (we
 * subtract 1 because the user specifies channels 1..K, and we number
 * them a la C as 0..K-1), and then store a -1, which is a special code
 * to mean all channels after the previous one.
 */

	static
	void
decode_channel_order_arg (

char *		ap		/* IN - argument to decode */
)
{
int		i;
int		min_chan;
int		max_chan;

#define ORDER	flag_channel_order
#define COUNT	flag_channel_order_count

	while (*ap NE '\0') {
		ap = cvint (ap, &min_chan, 1);
		if (*ap EQ '-') {
			++ap;
			ap = cvint (ap, &max_chan, 1);
			if ((*ap NE '\0') AND (*ap NE ',')) usage ();
			if (max_chan < min_chan) usage ();
			for (i = min_chan - 1; i < max_chan; i++) {
				if (COUNT >= MAX_CHANNELS) usage ();
				ORDER [COUNT++] = i;
			}
		}
		else {
			if ((*ap NE '\0') AND (*ap NE ',')) usage ();
			if (COUNT >= MAX_CHANNELS) usage ();
			ORDER [COUNT++] = min_chan - 1;
		}
		if (*ap EQ ',') {
			++ap;
			if (*ap EQ '\0') usage ();
		}
	}

#undef	ORDER
#undef	COUNT
}

/*
 * Convert a decimal number to an integer, enforcing a minimum value.
 */

	static
	char *
cvint (

char *		ap,		/* IN - string to convert */
int *		val_out,	/* OUT - channel number */
int		min		/* IN - minimum valid value */
)
{
char		c;
int		value;

	c = *ap;
	if ((c < '0') OR (c > '9')) {
		/* String MUST begin with a decimal number! */
		usage ();
	}
	value = c - '0';
	for (;;) {
		++ap;
		c = *ap;
		if ((c < '0') OR (c > '9')) break;
		value = 10 * value + (c - '0');
	}

	if (value < min) usage ();

	*val_out = value;

	return (ap);
}

/*
 * This routine prints out the proper usage and exits.
 */

static char *	arg_doc [] = {
"",
" -b T\tSet bottom title to T.",
" -c Ord\tDefine the display order of input channels down the page.  Ord",
"\t is a comma-separated sequence of items of the form:",
"\t N, or N-M, where N and M are positive decimal integers,",
"\t and N <= M.  These channel definitions are appended to any",
"\t obtained from the configuration file.",
" -f Cfg\tUse the specified configuration file.",
" -l\tPrint pages in landscape orientation.",
" -m S,F\tMagnify the vertical height of displayed strip number S by factor",
"\t F (relative to default height).",
" -p\tPrint pages in portrait orientation (this is the default unless",
"\t the configuration file specifies some other orientation).",
" -P N\tSpread the strip-chart across N pages.",
" -s X,Y\tSet paper size to X,Y (in postscript units of 1/72 of an inch).",
" -S S\tPlot data for strip S with \"scatter plotted\" dots instead of lines.",
" -t T\tSet top title to T.",
" -T foo\tActivate template foo.  Templates are defined in the configuration",
"\t file, and provides a set of options that plot the data in a specific",
"\t way.  This profixed a convenient way for the user to define several",
"\t meaningful views of the data that can be easily selected on the",
"\t command line without specifying dozens of dizzying little flags.  If",
"\t -T is not specified, a template named 'default' is used if it exists.",
" -v\tPrint version number and exit.",
" -x min:max",
"\tDisplay only those data records min through max.  The first data record",
"\t in the input file is considered to be record zero, and is displayed",
"\t at the X=0 position of the X axis (unless this argument causes some",
"\t other subrange of the records to be displayed).  Causes the X-axis to",
"\t be scaled accordingly.",
"",
NULL
};

	static
	void
usage (void)

{
char **		pp;
char *		p;

	(void) fprintf (stderr,
			"\nUsage: %s"
			" [-l]"
			" [-b BottomTitle]"
			" [-c channel_ordering]"
			" \\\n\t"
			" [-f ConfigFile]"
			" [-m StripNum,MagnificationFactor]"
			" \\\n\t"
			" [-p]"
			" [-P num_pages]"
			" [-s Xsize,Ysize]"
			" [-S StripNum]"
			" \\\n\t"
			" [-t TopTitle]"
			" [-T templateName]"
			" [-v]"
			" [-x xmin:xmax]"
			" <data-file"
			" \\\n\t"
			" >postscript-file\n",
			me);
	pp = &arg_doc [0];
	while ((p = *pp++) NE NULL) {
		(void) fprintf (stderr, "%s\n", p);
	}
	exit (1);
}

/*
 * Find and parse the configuration file.
 */

	static
	void
parse_config_file (void)

{
int		i, n;
FILE *		fp;
char *		newpath;
const char *	path;
const char *	homedir;

	newpath = NULL;

	path = flag_config_file_path;
	if (path NE NULL) {
		fp = fopen (path, "r");
		if (fp EQ NULL) {
			fprintf (stderr, "%s: %s - %s\n",
				 me, strerror (errno), path);
			exit (1);
		}
	}
	else {
		path = "./eeg.cfg";
		fp = fopen (path, "r");
		if (fp EQ NULL) {
			homedir = getenv ("HOME");
			n = strlen (homedir);
			n += 10;
			newpath = NEWA (n, char);
			strcpy (newpath, homedir);
			strcat (newpath, "/.eeg.cfg");
			path = newpath;
			fp = fopen (path, "r");
			if (fp EQ NULL) {
				fprintf (stderr, "%s: Cannot find config file.\n", me);
				exit (1);
			}
		}
	}

	yyin = fp;

	i = yyparse ();

	if (i NE 0) {
		fprintf (stderr,
			 "%s: Unrecoverable syntax error in config file %s\n",
			 me, path);
		exit (1);
	}

#if 0
	debug_print_parse_tree ();
#endif

	if (newpath NE NULL) {
		free (newpath);
	}
}

/*
 * Print out the config file parse tree for debugging purposes.
 */

	static
	void
debug_print_parse_tree (void)

{
struct SourceDefPnode *		sdp;
struct SourceDefItemPnode *	sdip;
struct VersionDefPnode *	vdp;
struct FieldDefPnode *		fdp;
struct LogicalFieldNamePnode *	lfnp;
struct PlotDefPnode *		pdp;
struct PlotOptionPnode *	popt;
struct ChanDefPnode *		cdp;
struct ChanDefOptionPnode *	cdop;

 if (root_pnode EQ NULL) {
  printf ("root_pnode is NULL!\n");
  return;
 }

 for (sdp = root_pnode -> srcdefs; sdp NE NULL; sdp = sdp -> next) {
  printf ("ID: \"%s\" {\n", sdp -> src_name -> text);
  for (sdip = sdp -> items; sdip NE NULL; sdip = sdip -> next) {
   switch (sdip -> type) {
   case SRCDEF_VERSION:
    vdp = sdip -> u.version_def;
    printf ("  Version: \"%s\" {\n", vdp -> version -> text);
    for (fdp = vdp -> field_defs; fdp NE NULL; fdp = fdp -> next) {
     printf ("    (");
     for (lfnp = fdp -> logical_names; lfnp NE NULL; lfnp = lfnp -> next) {
      printf (" \"%s\"", lfnp -> name -> text);
     }
     printf (")\n");
    }
    printf ("  }\n");
    break;

   case SRCDEF_PLOT:
    pdp = sdip -> u.plot_def;
    printf ("  Plot: \"%s\" {\n", pdp -> name -> text);
    for (popt = pdp -> options; popt NE NULL; popt = popt -> next) {
     switch (popt -> type) {
     case PLOTOPT_CHAN_DEF:
      for (cdp = popt -> u.chan_def; cdp NE NULL; cdp = cdp -> next) {
       printf ("    Chan: \"%s\" {\n", cdp -> name -> text);
       for (cdop = cdp -> options; cdop NE NULL; cdop = cdop -> next) {
	switch (cdop -> type) {
	case CHANOPT_DATA:
	 printf ("      Data: \"%s\"\n", cdop -> u.data_name -> text);
	 break;

	case CHANOPT_MAGNIFY:
	 printf ("      Mag:  %.17g\n", cdop -> u.magnification);
	 break;

	default:
	 abort ();
	}
       }
       printf ("    }\n");
      }
      break;

     case PLOTOPT_BOTTOM_TITLE:
      printf ("    Bottom_Title: \"%s\"\n", popt -> u.title -> text);
      break;

     case PLOTOPT_TOP_TITLE:
      printf ("    Top_Title: \"%s\"\n", popt -> u.title -> text);
      break;

     case PLOTOPT_ORIENTATION:
      printf ("    Orientation: %s\n",
	      popt -> u.landscape ? "Landscape" : "Portrait");
      break;

     default:
      abort ();
     }
    }
    printf ("  }\n");
    break;

    default:
    abort ();
   }
  }
  printf ("}\n");
 }
}

/*
 * Read in all of the leading "comment" lines until we get to lines
 * containing numeric data.  Scan these comment lines for the presence
 * of a special marker that indicates the "type" and "version" of the
 * data being plotted.  These are used to look up information in the
 * configuration file:
 *
 * 1. The "logical name" of each data field in the records of this data.
 *
 * 2. A set of named plot types, each of which can provide all of the
 *    other parameters and settings needed to produce that plot from
 *    this data.
 *
 * Lines are considered to be comments if their first character is
 * @, # or %.  Such lines are scanned for the special marker @EEG_TYPE@,
 * followed by two quoted strings: (a) the "type" of the data, and
 * (b) the "version".  The "type" might for example be the name of the
 * program that produced the data being plotted, while the "version"
 * might be used to handle different evolutionary versions of the
 * data produced by that one program.  For example, it would be common
 * for a newer version to include additional fields that do not appear
 * in earlier versions.  Tagging the data with its proper version let's
 * us to handle multiple in a direct manner.
 */

	static
	void
read_marker (

FILE *		fp	/* IN - stream to read header info from */
)
{
int		c;
int		count;
int		mask, match_mask;
char *		pattern;
char *		p;
struct Name *	name1;
struct Name *	name2;
unsigned short	table [UCHAR_MAX + 1];

#define	MARKER	"@EEG_TYPE@"

	pattern = MARKER;

	memset (table, 0, sizeof (table));
	mask = 1;
	for (p = pattern; *p NE '\0'; p++) {
		table [*p & 0xFF] |= mask;
		mask <<= 1;
	}
	match_mask = mask;

	count = 0;

	for (;;) {
		/* Read character of next line. */
		c = fgetc (fp);
		if (c < 0) break;

		if ((c NE '@') AND (c NE '#') AND (c NE '%')) {
			/* Header lines have been stripped. */
			ungetc (c, fp);
			break;
		}

		/* We have a "comment" line.  Scan it for our pattern. */
		mask = 0;
		while (c NE '\n') {
			mask = ((mask | 1) & table [c & 0xFF]) << 1;
			if ((mask & match_mask) NE 0) {
				/* Entire pattern has been found! */
				++count;
				if (count > 1) {
					fprintf (stderr,
						 "%s: Multiple "
						 MARKER
						 " markers in input file.\n",
						 me);
					exit (1);
				}

				/* Marker seen.  Read in the "data type" */
				/* and "version" as two quoted strings.  */

				name1 = read_quoted_string (fp);
				if (name1 NE NULL) {
					name2 = read_quoted_string (fp);
					if (name2 NE NULL) {

						process_type_and_version (
							name1,
							name2);
					}
				}

				mask = 0;
			}
			c = fgetc (fp);
			if (c < 0) break;
		}
		if (c < 0) break;
	}
}

/*
 * Read in a double-quoted string (possibly proceeded by whitespace).
 * Turn it into a "struct Name" and return it.
 * Return NULL if we run into something other than a quoted string, or if
 * we encounter end-of-line or end-of-file before we see the complete
 * quoted string.
 */

	static
	struct Name *
read_quoted_string (

FILE *		fp		/* IN - file to read quoted string from */
)
{
int		c;
char *		cp;
char *		endp;
struct Name *	name;
char		buf [128];

	/* Look for opening quote and reject any non-whitespace. */
	for (;;) {
		c = fgetc (fp);
		if (c EQ '"') break;
		if ((c NE ' ') AND (c NE '\t') AND (c NE '\f')) goto err;
	}

	/* Collect the guts of the string. */

	cp = &buf [0];
	endp = cp + sizeof (buf);

	for (;;) {
		c = fgetc (fp);
		if (c EQ '\\') {
			c = fgetc (fp);
		}
		else if (c EQ '"') break;

		if ((c EQ '\n') OR (c < 0)) goto err;
		if (cp < endp) {
			*cp = c;
		}
		++cp;
	}

	name = NULL;

	if (cp <= endp) {
		/* We have the complete string.  Now make the name. */
		name = make_name_from_buffer (buf, cp);
	}
	else {
		fprintf (stderr,
			 "%s: Quoted string too long after "
			 MARKER
			 " marker in data file.\n",
			 me);
	}

	return (name);

err:
	fprintf (stderr,
		 "%s: Expected two quoted strings after "
		 MARKER
		 " marker in data file.\n",
		 me);

	return (NULL);
}

/*
 * Process the given "data type" and "version" by looking them up in
 * the configuration file.
 *
 * We emit a warning if either of these is not found in the config file.
 *
 * We set the data_type variable to be the parse tree corresponding to
 * this type of data, and the data_version variable to be the parse tree
 * corresponding to the version.
 *
 * If the version is found, this routine also applies all of the
 * proper logical names to each data field (i.e., column).  Given one
 * of these logical names, we can then easily determine the phyisical
 * field/column number of that data.
 */

	static
	void
process_type_and_version (

struct Name *	tp,		/* IN - name of data type */
struct Name *	vp		/* IN - name of data version */
)
{
int				fidx;
struct SourceDefPnode *		sdp;
struct SourceDefItemPnode *	sdip;
struct VersionDefPnode *	vdp;
struct FieldDefPnode *		fdp;
struct LogicalFieldNamePnode *	lfnp;
struct Name *			fname;

	data_type	= NULL;
	data_version	= NULL;

	for (sdp = root_pnode -> srcdefs; sdp NE NULL; sdp = sdp -> next) {
		if (sdp -> src_name EQ tp) {
			data_type = sdp;
			break;
		}
	}

	if (data_type EQ NULL) {
		fprintf (stderr,
			 "%s: Warning: data type \"%s\" not found"
			 " in configuration file.\n",
			 me, tp -> text);
		return;
	}

	/* Now look for the version. */

	for (sdip = data_type -> items; sdip NE NULL; sdip = sdip -> next) {
		if (sdip -> type NE SRCDEF_VERSION) continue;
		vdp = sdip -> u.version_def;
		if (vp EQ vdp -> version) {
			data_version = vdp;
			break;
		}
	}

	if (data_version EQ NULL) {
		fprintf (stderr,
			 "%s: Warning: data version \"%s\" not found"
			 " in configuration file.\n",
			 me, vp -> text);
		return;
	}

	/* Apply logical names to each field. */
	fidx = 0;
	for (fdp = data_version -> field_defs; fdp NE NULL; fdp = fdp -> next) {
		if (fidx >= MAX_CHANNELS) {
			fprintf (stderr,
				 "%s: Too many logical channels named"
				 " in configuration file for type / version"
				 " \"%s\" \"%s\".\n",
				 me, tp -> text, vp -> text);
			exit (1);
		}
		/* Field number "fidx" receives all of the logical	*/
		/* names listed in fdp -> logical_names.  The first of	*/
		/* these is the "primary" logical name -- the one we	*/
		/* display when referring to this physical data column.	*/
		lfnp = fdp -> logical_names;
		if (lfnp NE NULL) {
			fname = lfnp -> name;
			field_name [fidx] = fname;
		}
		for (; lfnp NE NULL; lfnp = lfnp -> next) {
			fname = lfnp -> name;
			if (fname -> fldno < 0) {
				fname -> fldno = fidx;
			}
			else {
				fprintf (stderr,
					 "%s: Error in configuration file:"
					 " type / version \"%s\" \"%s\""
					 " multiply defines logical field"
					 " name \"%s\".\n",
					 me, tp -> text, vp -> text,
					 fname -> text);
			}
		}
		++fidx;
	}
	config_nfields = fidx;
}

/*
 * Read in all of the data lines.  Read each column of data into a
 * separate "rubber-band" buffer of data values.  Verify that each
 * line has the same number of data fields.
 */

	static
	void
read_data (

FILE *		fp
)
{
int		i;
int		c;
int		fidx;
int		record;
struct Datum	datum;

	/* We don't know how many fields there are yet... */
	nfields = -1;

	record = 0;
	for (;;) {
		c = fgetc (fp);
		if (c < 0) break;
		if ((c EQ '@') OR (c EQ '#') OR (c EQ '%')) {
			/* Ignore comment lines. */
			for (;;) {
				c = fgetc (fp);
				if (c < 0) break;
				if (c EQ '\n') break;
			}
			if (c < 0) break;
			continue;
		}
		else {
			ungetc (c, fp);
		}
		fidx = 0;
		i = read_datum (fp, &datum);
		/* -1 ==> EOF, 0 ==> EOL, 1 ==> got a datum. */
		while (i > 0) {
			if (fidx >= MAX_CHANNELS) {
				fprintf (stderr,
					 "%s: Too many data fields!"
					 "  (Max is %d)\n",
					 me, MAX_CHANNELS);
				exit (1);
			}
			append_datum (&field_dbufs [fidx], &datum);
			++fidx;
			i = read_datum (fp, &datum);
		}
		if (i < 0) {
			/* End of file. */
			if (fidx NE 0) {
				fprintf (stderr,
					 "%s: EOF in middle of data record!\n",
					 me);
				exit (1);
			}
			if (nfields < 0) {
				nfields = 0;
			}
			break;
		}
		++record;
		if (nfields < 0) {
			nfields = fidx;
		}
		else if (fidx NE nfields) {
			fprintf (stderr,
				 "%s: Record %d has %d fields instead of %d.\n",
				 me, record, fidx, nfields);
			exit (1);
		}
	}
	nrecords = record;
}

/*
 * Now that we know exactly how many fields and records we have,
 * put all of the data for each field into a single linear array.
 */

	static
	void
put_data_into_arrays (void)

{
int		j, count;
struct DBuf *	bufp;
struct Datum *	array;

	for (j = 0; j < nfields; j++) {
		bufp = field_dbufs [j];
		field_dbufs [j] = NULL;
		if (bufp EQ NULL) abort ();

		linearize_dbuf (bufp, &array, &count);

		if (count NE nrecords) abort ();

		field_data [j] = array;

		free_dbuf (bufp);
	}
}

/*
 * Process all of the options for the given plot "template".  This is
 * just a name that we look up in the configuration file, and it gives
 * us a bunch of options to process.  (These options can be overridden
 * by the command line options, so we process them first.)
 */

	static
	void
process_selected_plot (

const char *	template_name	/* IN - name of plot template to use */
)
{
struct Name *			tname;
struct SourceDefItemPnode *	sdip;
struct PlotDefPnode *		pdp;

	option_bottom_title		= NULL;
	option_chan_def_count		= 0;
	option_top_title		= NULL;

	if (data_version NE NULL) {
		if (nfields NE config_nfields) {
			fprintf (stderr,
				 "%s: Warning -- data file contains %d fields"
				 " but config file specifies %d fields for"
				 " this type and version of data.\n",
				 me, nfields, config_nfields);
		}
	}

	if (data_type EQ NULL) {
		/* We have no plot definitions to look for. */
		if (template_name NE NULL) {
			fprintf (stderr,
				 "%s: Data type is unknown, making it"
				 " impossible to interpret template \"%s\".\n",
				 me, template_name);
			exit (1);
		}
		return;
	}

	/* Find the given template name (or "default" if none given).	*/

	if (template_name NE NULL) {
		tname = make_name (template_name);
	}
	else {
		tname = make_name ("default");
	}

	for (sdip = data_type -> items; sdip NE NULL; sdip = sdip -> next) {
		if (sdip -> type NE SRCDEF_PLOT) continue;
		pdp = sdip -> u.plot_def;
		if (pdp -> name EQ tname) {
			process_option_list (pdp -> options);
			break;
		}
	}

	/* Use options from config file that were not specified	*/
	/* on the command line.					*/

	if (flag_bottom_title EQ NULL) {
		flag_bottom_title = option_bottom_title;
	}
	if (flag_top_title EQ NULL) {
		flag_top_title = option_top_title;
	}
	if ((NOT flag_pagemode_given) AND (option_pagemode_given)) {
		flag_pagemode = option_pagemode;
	}
}

/*
 * Process all of the options in the given list.
 */

	static
	void
process_option_list (

struct PlotOptionPnode * options	/* IN - options to process */
)
{
int				i, j;
int				mag_given;
int				scatter;
int				force_linear;
struct PlotOptionPnode *	op;
struct ChanDefPnode *		cdp;
struct ChanDefOptionPnode *	cdop;
struct Name *			cname;
struct Name *			fname;
struct ChannelDef *		cd;
double				mag, linear_min, linear_max;

	for (op = options; op NE NULL; op = op -> next) {
		switch (op -> type) {
		case PLOTOPT_CHAN_DEF:
			cdp = op -> u.chan_def;
			cname = cdp -> name;

			/* Process the channel definition options. */
			force_linear	= 0;
			linear_min	= 0.0;
			linear_max	= 0.0;
			mag		= 1.0;
			mag_given	= FALSE;
			fname		= NULL;
			scatter		= FALSE;

			for (cdop = cdp -> options; cdop NE NULL; cdop = cdop -> next) {
				switch (cdop -> type) {
				case CHANOPT_DATA:
					if (fname NE NULL) {
						fprintf (stderr,
							 "%s: Data multiply"
							 " defined for channel"
							 " \"%s\".\n",
							 me,
							 cname -> text);
						exit (1);
					}
					fname = cdop -> u.data_name;
					break;

				case CHANOPT_LINEAR:
					if (force_linear) {
						fprintf (stderr,
							 "%s: Linear scale"
							 " multiply defined"
							 " for channel"
							 " \"%s\".\n",
							 me,
							 cname -> text);
						exit (1);
					}
					linear_min = cdop -> u.linear.min;
					linear_max = cdop -> u.linear.max;
					if (linear_min >= linear_max) {
						fprintf (stderr,
							 "%s: Invalid linear"
							 " scale for channel"
							 " \"%s\".\n",
							 me,
							 cname -> text);
						exit (1);
					}
					force_linear = TRUE;
					break;

				case CHANOPT_MAGNIFY:
					if (mag_given) {
						fprintf (stderr,
							 "%s: Magnification"
							 " multiply defined"
							 " for channel"
							 " \"%s\".\n",
							 me,
							 cname -> text);
						exit (1);
					}
					mag_given = TRUE;
					mag = cdop -> u.magnification;
					break;

				case CHANOPT_SCATTER:
					if (scatter) {
						fprintf (stderr,
							 "%s: Scatter"
							 " specified more than"
							 " once for channel"
							 " \"%s\".\n",
							 me,
							 cname -> text);
						exit (1);
					}
					scatter = TRUE;
					break;

				default:
					abort ();
				}
			}

			if (fname EQ NULL) {
				/* Did not specify "Data", so presume	*/
				/* that the channel "name" is also the	*/
				/* logical field name to plot in this	*/
				/* channel.				*/
				fname = cname;
			}

			j = fname -> fldno;
			if (j < 0) {
				/* Issue a warning and ignore this channel. */
				fprintf (stderr,
					 "%s: No field named \"%s\" is known"
					 " for this type and version of"
					 " data.\n",
					 me, fname -> text);
				break;
			}
			if (j >= nfields) {
				/* Issue a warning and ignore this channel. */
				fprintf (stderr,
					 "%s: The field named \"%s\" is"
					 " defined for this type and version"
					 " of data, but the data records do"
					 " not contain enough fields to"
					 " actually provide this data."
					 "  Ignoring this channel.\n",
					 me, fname -> text);
				break;
			}

			/* Define which physical column to plot	*/
			/* in the next strip/channel, and set	*/
			/* its default title.			*/
			i = option_chan_def_count++;
			if (i >= MAX_CHANNELS) abort ();

			cd = &option_chan_defs [i];

			cd -> fldno		= j;
			cd -> title		= cname -> text;
			cd -> mag_factor	= mag;
			cd -> linear_min	= linear_min;
			cd -> linear_max	= linear_max;
			cd -> force_linear	= force_linear;
			cd -> scatter		= scatter;
			break;

		case PLOTOPT_BOTTOM_TITLE:
			option_bottom_title = op -> u.title -> text;
			break;

		case PLOTOPT_TOP_TITLE:
			option_top_title = op -> u.title -> text;
			break;

		case PLOTOPT_ORIENTATION:
			if (op -> u.landscape) {
				option_pagemode = PAGEMODE_LANDSCAPE;
			}
			else {
				option_pagemode = PAGEMODE_PORTRAIT;
			}
			option_pagemode_given = TRUE;
			break;

		default:
			abort ();
		}
	}
}

/*
 * If the user has specified "-x xmin,xmax", then he wants us to "zoom in"
 * on just these records.  If not, then we display all records present.
 */

	static
	void
isolate_selected_data (void)

{
	if (xmin < 0) {
		/* The user hasn't specified -x.  Display all data! */
		xmin = 0;
		xmax = nrecords - 1;
	}
	else if (xmax > nrecords) {
		fprintf (stderr,
			 "%s: -x specifies data out of range\n"
			 "Maximum record number is %d\n",
			 me, nrecords);
		exit (1);
	}
	if ((0 > xmin) OR (xmin >= xmax) OR (xmax > nrecords)) abort ();
}

/*
 * Define the number of "strips / channels" of data that will be displayed.
 * For each such strip, determine which data field is to be displayed.
 */

	static
	void
define_strips (void)

{
int			i, j, k, n;
const char *		title;
struct StripInfo *	sip;
struct ChannelDef *	cdp;
struct Name *		name;
double			factor;
char 			buffer [32];

	k = 0;

	if (flag_channel_order_count > 0) {
		/* The user has explicitly chosen the order in which */
		/* to display the various data fields. */
		n = flag_channel_order_count;
		for (i = 0; i < n; i++) {
			j = flag_channel_order [i];
			if ((j < 0) OR (j >= nfields)) {
				fprintf (stderr,
					 "%s: Reference to non-existent field\n",
					 me);
				exit (1);
			}
			sip = &strip_info [k];
			sip -> mag_factor	= 1.0;
			sip -> field		= j;
			/* FIXME - force linear is currently inaccessible */
			/* from the command line.			  */
			sip -> linear_min	= 0.0;
			sip -> linear_max	= 0.0;
			sip -> force_linear	= FALSE;
			sip -> title		= NULL;
			sip -> scatter		= flag_scatter_plot [i];
			++k;
		}
	}
	else if (option_chan_def_count > 0) {
		/* The user has not provided a channel order on the	*/
		/* command line, but one has been defined via the	*/
		/* configuration file.					*/
		n = option_chan_def_count;
		for (i = 0; i < n; i++) {
			cdp = &option_chan_defs [i];
			j = cdp -> fldno;
			/* This should have been verified eariler... */
			if ((j < 0) OR (j >= nfields)) abort ();

			sip = &strip_info [k];

			sip -> mag_factor	= cdp -> mag_factor;
			sip -> field		= j;
			sip -> linear_min	= cdp -> linear_min;
			sip -> linear_max	= cdp -> linear_max;
			sip -> force_linear	= cdp -> force_linear;
			sip -> title		= cdp -> title;
			sip -> scatter		= cdp -> scatter;
			++k;
		}
		nstrips = k;
		return;
	}
	else {
		/* Display all fields in the order that they appear in	*/
		/* the data file.					*/
		for (i = 0; i < nfields; i++) {
			sip = &strip_info [k];
			sip -> mag_factor	= 1.0;
			sip -> field		= i;
			sip -> linear_min	= 0.0;
			sip -> linear_max	= 0.0;
			sip -> force_linear	= FALSE;
			sip -> title		= NULL;
			sip -> scatter		= flag_scatter_plot [i];
			++k;
		}
	}
	nstrips = k;

	/* Any strips that do not yet have titles, grab the title */
	/* from the field information, or manufacture a title.	  */
	for (i = 0; i < k; i++) {
		sip = &strip_info [i];
		title = sip -> title;
		if (title NE NULL) continue;

		j = sip -> field;
		if ((j < 0) OR (j >= nfields)) abort ();

		name = field_name [j];
		if (name NE NULL) {
			sip -> title = name -> text;
		}
		else {
			/* Construct a synthetic field title. */
			sprintf (buffer, "Field %d", j + 1);
			title = mystrdup (buffer);
			sip -> title = title;
		}
	}

	/* Apply magnification factors. */

	n = flag_mag_count;
	for (i = 0; i < n; i++) {
		j	= flag_mag_strip_num [i];
		factor	= flag_mag_factor [i];
		if ((j < 1) OR (j > nstrips)) {
			fprintf (stderr,
				 "%s: Mag factor for invalid strip number %d\n",
				 me, j);
			exit (1);
		}
		if (factor <= 0.0) {
			fprintf (stderr, "%s: Mag factors must be positive\n", me);
			exit (1);
		}
		strip_info [j - 1].mag_factor *= factor;
	}
}

/*
 * Partition the page into various strips, some of which are optional:
 *
 *	1. top title
 *	2. one strip per channel
 *	3. bottom title
 *
 * Determine the Ymin / Ymax coordinates for each of these strips,
 * and record all of this information.
 */

	static
	void
partition_page_into_horizontal_strips (void)

{
int				i;
const struct PaperSize *	ps;
struct StripInfo *		sip;
double				height;
double				units, units2;
double				alpha;
double				ymin, ymax;

	/* First tally the total number of units high. */
	units = 0.0;
#if 0
	if (flag_top_title NE NULL) {
		/* The top title is 1/4 default height of a data strip. */
		units += 0.25;
		units += 0.0625;
	}
#endif
	for (i = 0; i < nstrips; i++) {
		units += strip_info [i].mag_factor;
		/* Add 1/8 of a strip spacing. */
		units += 0.125;
	}
#if 0
	if (flag_bottom_title NE NULL) {
		/* The bottom title is 1/4 default height of a data strip. */
		units += 0.0625;
		units += 0.25;
	}
#endif

	if (units <= 0.0) {
		fprintf (stderr, "%s: Nothing to plot\n", me);
		exit (1);
	}

	ps = flag_papersize;
	if (flag_pagemode EQ PAGEMODE_PORTRAIT) {
		height = ps -> y;
	}
	else {
		height = ps -> x;
	}

	/* We need to leave a margin at top and bottom. */
	ymin	= MARGIN;
	ymax	= height - MARGIN;

	if (flag_top_title NE NULL) {
		/* Top title is 1/4". */
		layout.top_title_top_y = ymax;
		ymax -= 18.0;
		layout.top_title_bot_y = ymax;
	}

	/* Leave space for ticks along the top edge of the top-most strip. */
	ymax -= 9.0;

	if (flag_bottom_title NE NULL) {
		layout.bottom_title_bot_y = ymin;
		ymin += 18.0;
		layout.bottom_title_top_y = ymin;
	}

	/* Leave space for xscale numbering and ticks along the bottom	*/
	/* edge of the bottom-most strip.				*/
	layout.bottom_scale_ymin = ymin;
	ymin += 12.0;
	layout.bottom_scale_ymax = ymin;
	ymin += 9.0;

	if (ymin >= ymax) abort ();	/* tiny paper! */

	units2 = 0.0;
#if 0
	if (flag_top_title NE NULL) {
		alpha = units2 / units;
		layout.top_title_top_y = (1.0 - alpha) * ymax + alpha * ymin;
		units2 += 0.25;
		alpha = units2 / units;
		layout.top_title_bot_y = (1.0 - alpha) * ymax + alpha * ymin;
		units2 += 0.0625;
	}
#endif
	for (i = 0; i < nstrips; i++) {
		sip = &strip_info [i];
		units2 += 0.03125;
		alpha = units2 / units;
		sip -> scale_ymax = (1.0 - alpha) * ymax + alpha * ymin;
		units2 += 0.03125;
		alpha = units2 / units;
		sip -> top_y = (1.0 - alpha) * ymax + alpha * ymin;
		units2 += sip -> mag_factor;
		alpha = units2 / units;
		sip -> bot_y = (1.0 - alpha) * ymax + alpha * ymin;
		units2 += 0.03125;
		alpha = units2 / units;
		sip -> scale_ymin = (1.0 - alpha) * ymax + alpha * ymin;
		units2 += 0.03125;
	}
#if 0
	if (flag_bottom_title NE NULL) {
		units2 += 0.0625;
		alpha = units2 / units;
		layout.bottom_title_top_y = (1.0 - alpha) * ymax + alpha * ymin;
		units2 += 0.25;
		alpha = units2 / units;
		layout.bottom_title_bot_y = (1.0 - alpha) * ymax + alpha * ymin;
	}
#endif
}

/*
 * Partition the page into various vertical strips:
 *
 * 1. Channel title boxes
 * 2. Channel scale label boxes
 * 3. Channel data plotting boxes
 */

	static
	void
partition_page_into_vertical_strips (void)

{
const struct PaperSize * ps;
double			width;
double			left, right;

	ps = flag_papersize;
	width = (flag_pagemode EQ PAGEMODE_PORTRAIT) ? ps -> x : ps -> y;

	/* Leave margins on left and right. */
	left = MARGIN;
	right = width - MARGIN;

	layout.xmin	= left;
	layout.xmax	= right;

	/* We reserve horizontal space for the following:		*/
	/*								*/
	/* 1.	Data channel labels (1/2 inch).				*/
	/* 2.	Numeric scale values (1/2 inch).			*/
	/* 3.	External tick marks (1/16 inch).			*/
	/* 4.	Gap before first plotted datum (1/16 inch).		*/
	/* 5.	The data plotting area (whatever is left).		*/
	/* 6.	Gap before channel box end (1/16 inch).			*/
	/* 7.	External tick marks (1/16 inch).			*/
	/* 8.	Numeric scale values (1/2 inch).			*/

	layout.chan_label_xmin		= left;
	left += 36.0;
	layout.chan_label_xmax		= left;
	layout.left_scale_xmin		= left;
	left += 36.0;
	layout.left_scale_xmax		= left;
	layout.left_tick_xmin		= left;
	left += 4.5;
	layout.left_tick_xmax		= left;
	layout.plot_area_xmin		= left;
	left += 4.5;
	layout.data_xmin		= left;

	layout.right_scale_xmax		= right;
	right -= 36.0;
	layout.right_scale_xmin		= right;
	layout.right_tick_xmax		= right;
	right -= 4.5;
	layout.right_tick_xmin		= right;
	layout.plot_area_xmax		= right;
	right -= 4.5;
	layout.data_xmax		= right;

	if (left >= right) abort ();	/* tiny paper! */
}

/*
 * Determine how we will scale the X-axis.
 *
 * This is pretty simple -- unless we are splitting the plot across
 * multiple pages.
 */

	static
	void
choose_x_axis (void)

{
int		n;
int		pages;
double		pow10;

	pages = flag_numpages;

	if (pages < 1) abort ();

	n = xmax - xmin + 1;

	if (pages EQ 1) {
		for (pow10 = 1; ; pow10 *= 10.0) {
			if (pow10 >= 1000000000.0) {
				fprintf (stderr, "%s: Too many records!\n", me);
				exit (1);
			}
			if (n / pow10 <= 10) {
				scale_xtics = pow10;
				break;
			}
			if (n / (2 * pow10) <= 10) {
				scale_xtics = 2 * pow10;
				break;
			}
			if (n / (5 * pow10) <= 10) {
				scale_xtics = 5 * pow10;
				break;
			}
#if 0
			if (n <= 5.0 * pow10) {
				scale_xmax = 5.0 * pow10;
				scale_xtics = pow10;
				break;
			}
			else if (n <= 10.0 * pow10) {
				scale_xmax = 10.0 * pow10;
				scale_xtics = pow10;
				break;
			}
			else if (n <= 20.0 * pow10) {
				scale_xmax = 20.0 * pow10;
				scale_xtics = 2.0 * pow10;
				break;
			}
#endif
		}
		scale_xmin = floor (xmin / scale_xtics) * scale_xtics;
		scale_xmax = ceil  (xmax / scale_xtics) * scale_xtics;
		return;
	}

	fprintf (stderr, "%s: Multi-page plots not yet implemented!\n", me);
	exit (1);
}

/*
 * Pick the Y-axis format and range for each strip being displayed.
 */

	static
	void
choose_y_axes (void)

{
int			i, n;
struct StripInfo *	sip;

	n = nstrips;
	for (i = 0; i < n; i++) {
		sip = &strip_info [i];
		choose_y_axis (sip);
	}
}

/*
 * Determine the format and range of the Y axis for the given strip.
 */

	static
	void
choose_y_axis (

struct StripInfo *	sip	/* IN/OUT - strip to choose Y axis for */
)
{
int			i, j;
int			top_item, bot_item;
const struct Datum *	data;
const struct Datum *	dp;
int			_pinf, _plus, _zero, _minus, _minf, _nan;
double			y, top, bot, middle;
double			pmax, pmin, mmax, mmin;

	/* Scan data in this strip to determine the presence of	*/
	/* special data values (+inf, -inf, nan, zero), and the	*/
	/* presence / range of positive and/or negative values.	*/

	j    = sip -> field;
	data = field_data [j];

	_pinf = _plus = _zero = _minus = _minf = _nan = FALSE;
	pmax = pmin = mmax = mmin = 0.0;

	for (i = xmin; i <= xmax; i++) {
		dp = &data [i];
		switch (dp -> type) {
		case DATUM_TYPE_VALUE:
			break;

		case DATUM_TYPE_PINF:
			_pinf = TRUE;
			continue;

		case DATUM_TYPE_MINF:
			_minf = TRUE;
			continue;

		case DATUM_TYPE_NAN:
			_nan = TRUE;
			continue;

		default:
			abort ();
		}
		/* We have a numeric value. */
		y = dp -> value;
		if (sip -> force_linear) {
			/* Treat offscale data as -inf or +inf. */
			if (y < sip -> linear_min) {
				_minf = TRUE;
				continue;
			}
			else if (y > sip -> linear_max) {
				_pinf = TRUE;
				continue;
			}
		}
		if (y EQ 0.0) {
			_zero = TRUE;
		}
		else if (y > 0) {
			if (NOT _plus) {
				pmax = pmin = y;
				_plus = TRUE;
			}
			else {
				if (y > pmax) pmax = y;
				if (y < pmin) pmin = y;
			}
		}
		else {
			y = fabs (y);
			if (NOT _minus) {
				mmax = mmin = y;
				_minus = TRUE;
			}
			else {
				if (y > mmax) mmax = y;
				if (y < mmin) mmin = y;
			}
		}
	}

	/* First determine which "bands" we will have: +inf, plus,	*/
	/* zero, minus, -inf, and nan.  To simplify things, we never	*/
	/* omit unused items if they are between two items that do	*/
	/* exist.							*/

	if (_pinf)	 { top_item = 0; }
	else if (_plus)  { top_item = 1; }
	else if (_zero)  { top_item = 2; }
	else if (_minus) { top_item = 3; }
	else if (_minf)  { top_item = 4; }
	else if (_nan)   { top_item = 5; }
	else { abort (); }

	if (_nan)	 { bot_item = 5; }
	else if (_minf)	 { bot_item = 4; }
	else if (_minus) { bot_item = 3; }
	else if (_zero)	 { bot_item = 2; }
	else if (_plus)	 { bot_item = 1; }
	else if (_pinf)	 { bot_item = 0; }
	else { abort (); }

	if (top_item > bot_item) abort ();

	sip -> top_band = top_item;
	sip -> bot_band = bot_item;

	/* We are displaying every item from top_item through bot_item. */

	if (_plus) {
		scale_numeric_band (pmin, pmax, sip, TRUE);
	}
	else {
		empty_numeric_band (&(sip -> pband));
	}
	if (_minus) {
		scale_numeric_band (mmin, mmax, sip, FALSE);
	}
	else {
		empty_numeric_band (&(sip -> mband));
	}
	if (_plus AND _minus) {
		/* Displaying both sign bands.  Also display zero line! */
		_zero = TRUE;
	}

	if ((top_item <= 1) AND (bot_item EQ 2) AND
	    (sip -> pband.logscale EQ 0) AND (sip -> pband.minval <= 0)) {
		/* We are displaying both the positive band and the zero */
		/* line, but the positive band includes zero in its range. */
		/* Drop the display of a separate zero line. */
		bot_item	= 1;
		sip -> bot_band = bot_item;
	}

	if ((top_item EQ 2) AND (bot_item EQ 2)) {
		/* Displaying only the zero line!  We change this so	*/
		/* that we display the interval [-1,+1] on the		*/
		/* "positive" band.					*/
		top_item = 1;
		bot_item = 1;
		sip -> top_band	= top_item;
		sip -> bot_band	= bot_item;
		sip -> pband.minval = -1.0;
		sip -> pband.maxval = +1.0;
		sip -> pband.ytics  = 1.0;
	}

	sip -> pinf_pos		= -1.0;
	sip -> pinf_label_top	= -1.0;
	sip -> pinf_label_bot	= -1.0;
	sip -> plus_top_pos	= -1.0;
	sip -> plus_bot_pos	= -1.0;
	sip -> plus_label_top	= -1.0;
	sip -> plus_label_bot	= -1.0;
	sip -> zero_pos		= -1.0;
	sip -> zero_label_top	= -1.0;
	sip -> zero_label_bot	= -1.0;
	sip -> minus_top_pos	= -1.0;
	sip -> minus_bot_pos	= -1.0;
	sip -> minus_label_top	= -1.0;
	sip -> minus_label_bot	= -1.0;
	sip -> minf_pos		= -1.0;
	sip -> minf_label_top	= -1.0;
	sip -> minf_label_bot	= -1.0;
	sip -> nan_pos		= -1.0;
	sip -> nan_label_top	= -1.0;
	sip -> nan_label_bot	= -1.0;

	top = 1.0;
	bot = 0.0;
	if ((top_item <= 0) AND (0 <= bot_item)) {
		sip -> pinf_pos		= top;
		sip -> pinf_label_top	= top + 0.03125;
		sip -> pinf_label_bot	= top - 0.03125;
		top -= 0.0625;
	}
	if ((top_item <= 5) AND (5 <= bot_item)) {
		sip -> nan_pos		= bot;
		sip -> nan_label_top	= bot + 0.03125;
		sip -> nan_label_bot	= bot - 0.03125;
		bot += 0.0625;
	}
	if ((top_item <= 4) AND (4 <= bot_item)) {
		sip -> minf_pos		= bot;
		sip -> minf_label_top	= bot + 0.03125;
		sip -> minf_label_bot	= bot - 0.03125;
		bot += 0.0625;
	}

	if ((top_item <= 1) AND (3 <= bot_item)) {
		/* Displaying both + and - bands. */
		middle = 0.5 * (top + bot);
		sip -> plus_label_top	= top + 0.03125;
		sip -> plus_top_pos	= top;
		sip -> plus_bot_pos	= middle + 0.0625;
		sip -> plus_label_bot	= middle + 0.03125;

		sip -> zero_label_top	= middle + 0.03125;
		sip -> zero_pos		= middle;
		sip -> zero_label_bot	= middle - 0.03125;

		sip -> minus_label_top	= middle - 0.03125;
		sip -> minus_top_pos	= middle - 0.0625;
		sip -> minus_bot_pos	= bot;
		sip -> minus_label_bot	= bot - 0.03125;

		return;
	}

	if (top_item <= 1) {
		if ((top_item <= 2) AND (2 <= bot_item)) {
			/* Displaying +band, zero line. */
			sip -> zero_pos		= bot;
			sip -> zero_label_top	= bot + 0.03125;
			sip -> zero_label_bot	= bot - 0.03125;
			bot += 0.0625;
		}
		sip -> plus_label_top	= top + 0.03125;
		sip -> plus_top_pos	= top;
		sip -> plus_bot_pos	= bot;
		sip -> plus_label_bot	= bot - 0.03125;
	}
	else {
		/* top_item >= 2 */
		if ((top_item <= 2) AND (2 <= bot_item)) {
			/* Displaying zero line, -band. */
			sip -> zero_pos		= top;
			sip -> zero_label_top	= top + 0.03125;
			sip -> zero_label_bot	= top - 0.03125;
			top -= 0.0625;
		}
		sip -> minus_label_top	= top + 0.03125;
		sip -> minus_top_pos	= top;
		sip -> minus_bot_pos	= bot;
		sip -> minus_label_bot	= bot - 0.03125;
	}
}

/*
 * Determine the Y-axis scaling of a all-positive or all-negative "band"
 * of numbers whose min and max magnitude are given.
 */

	static
	void
scale_numeric_band (

double		minmag,		/* IN - minimum numeric magnitude */
double		maxmag,		/* IN - maximum numeric magnitude */
struct StripInfo * sip,		/* IN - strip we are scaling a band for */
int		pband		/* IN - positive band? */
)
{
int		i, j, n;
int		minlog, maxlog;
struct BScale *	bscale;
double		minmant;

	if (pband) {
		bscale = &(sip -> pband);
	}
	else {
		bscale = &(sip -> mband);
	}

	memset (bscale, 0, sizeof (*bscale));

	if (sip -> force_linear) {
		/* Force a linear scale. */
		bscale -> logscale = FALSE;
		if (pband) {
			if (sip -> linear_max <= 0.0) {
				abort ();
			}
			bscale -> maxval = sip -> linear_max;
			bscale -> minval = sip -> linear_min;
			if (bscale -> minval < 0.0) {
				bscale -> minval = 0.0;
			}
		}
		else {
			if (sip -> linear_min >= 0.0) {
				abort ();
			}
			bscale -> maxval = - (sip -> linear_min);
			bscale -> minval = - (sip -> linear_max);
			if (bscale -> minval < 0.0) {
				bscale -> minval = 0.0;
			}
		}
		/* FIXME */
		bscale -> ytics = (bscale -> maxval - bscale -> minval) / 5.0;
	}
	else if (maxmag / minmag <= 20.0) {
		/* We should use a linear scale. */
		bscale -> logscale = FALSE;
		if (maxmag <= minmag) {
			/* Only 1 value! */
			bscale -> minval = 0.0;
			bscale -> maxval = 2 * minmag;
			bscale -> ytics = minmag;
			return;
		}
		minmant = exp10 (floor (log10 (minmag)));
		for (i = 2; i <= 10; i++) {
			if (minmant * i >= minmag) break;
		}
		--i;
		j = ceil (maxmag / minmant);
		n = j - i + 1;
		bscale -> minval = i * minmant;
		bscale -> maxval = j * minmant;
		if (n <= 5) {
			bscale -> ytics = minmant;
		}
		else if (n <= 10) {
			bscale -> ytics = 2 * minmant;
		}
		else if (n <= 20) {
			bscale -> ytics = 5 * minmant;
		}
		else {
			bscale -> ytics = 10 * minmant;
		}
	}
	else {
		/* Determine power of 10 for lower end of scale. */
		minlog = floor (log10 (minmag));

		maxlog = ceil (log10 (maxmag));

		bscale -> logscale	= TRUE;
		bscale -> minpow	= minlog;
		bscale -> maxpow	= maxlog;
		n = maxlog - minlog;
		if (n <= 5) {
			bscale -> powtics = 1;
		}
		else if ((n % 5) EQ 0) {
			bscale -> powtics = n / 5;
		}
		else if ((n % 4) EQ 0) {
			bscale -> powtics = n / 4;
		}
		else if ((n % 3) EQ 0) {
			bscale -> powtics = n / 3;
		}
		else  {
			bscale -> powtics = n / 5;
		}
	}
}

/*
 * Scale a numeric band for which there are no data points.
 */

	static
	void
empty_numeric_band (

struct BScale *	bscale
)
{
	memset (bscale, 0, sizeof (*bscale));
	bscale -> logscale = FALSE;
	bscale -> minval	= 0.0;
	bscale -> maxval	= 1.0;
	bscale -> ytics		= 0.5;
}

/*
 * Emit the first few lines of Postscript -- the "comments" section.
 */

	static
	void
emit_comments_section (void)

{
time_t		now;
char *		p;
const char *	orientation;
const struct PaperSize * ps;
char		tbuf [32];

	now = time (NULL);

	ctime_r (&now, tbuf);
	p = strchr (tbuf, '\n');
	if (p NE NULL) {
		*p = '\0';
	}

	orientation = (flag_pagemode EQ PAGEMODE_PORTRAIT)
			? "Portrait" : "Landscape";

	ps = flag_papersize;

	printf ("%%!PS-Adobe-3.0\n"
		"%%%%Creator: EEG Version %s\n"
		"%%%%Copyright: Copyright (c) 1998, 2009 by David M. Warme."
		"  All rights reserved.\n"
		"%%%%CreationDate: %s\n"
		"%%%%LanguageLevel: 1\n"
		"%%%%DocumentNeededResources: font Times-Roman\n"
		"%%%%DocumentSuppliedResources:"
		" procset StripChartProcs 1.0 0\n"
		"%%%%Orientation: %s\n"
		"%%%%Pages: %d\n"
		"%%%%DocumentMedia: %s %g %g 0 () ()\n"
		"%%%%EndComments\n",
		PACKAGE_VERSION_STRING,
		tbuf, orientation, flag_numpages,
		ps -> name,
		ps -> x,
		ps -> y);
}

/*
 * Emit the "prolog" section of the Postscript file.
 */

	static
	void
emit_prolog_section (void)

{
	fwrite (ps_prolog, 1, sizeof (ps_prolog), stdout);
}

/*
 * Emit the "setup" section of the Postscript file.
 */

	static
	void
emit_setup_section (void)

{
	printf ("%%%%BeginSetup\n"
		"%%%%PaperSize: %s\n"
		"\n"
		"%%EndSetup\n",
		flag_papersize -> name);
}

/*
 * Emit a simple version of the plots.
 */

	static
	void
emit_plot_page (void)

{
int				i;
struct StripInfo *		sip;

	printf ("%%%%Page: 1 1\n");
	printf ("StripChartDict begin\n");
	if (flag_pagemode EQ PAGEMODE_LANDSCAPE) {
		printf ("%% Transform coordinates for landscape\n"
			" %f 0 translate\n"
			" 90 rotate\n",
			flag_papersize -> x);
	}

	if (flag_top_title NE NULL) {
		printf ("\n %% Top title\n");
		printf (" %f %f %f %f /Helvetica 15 ",
			layout.xmin, layout.top_title_bot_y,
			layout.xmax, layout.top_title_top_y);
		emit_postscript_string (flag_top_title);
		printf (" ShowTextInBox\n");
	}
	for (i = 0; i < nstrips; i++) {
		sip = &strip_info [i];
		printf ("\n %% Strip %d\n", i);
#if 0
		box (layout.xmin, layout.xmax, sip -> bot_y, sip -> top_y);
#endif
		plot_strip_data (sip);
	}
	plot_x_ticks_and_scale ();
	if (flag_bottom_title NE NULL) {
		printf ("\n %% Bottom title\n");
		printf (" %f %f %f %f /Helvetica 15 ",
			layout.xmin, layout.bottom_title_bot_y,
			layout.xmax, layout.bottom_title_top_y);
		emit_postscript_string (flag_bottom_title);
		printf (" ShowTextInBox\n");
	}
	printf (" showpage\n");
	printf ("end\n");
}

/*
 * Draw the data plot for a single strip channel.
 */

	static
	void
plot_strip_data (

struct StripInfo *	sip
)
{
int			i, j;
int			top_band, bot_band;
int			nseg;
const struct Datum *	data;
const struct Datum *	dp;
const char *		op;
double			x, y, y1, y2, y3, y4;
double			alpha, denom;
double			top_y, bot_y, width;

	/* Display label for this data channel. */
	if (sip -> title NE NULL) {
		printf (" %% Label for data channel\n");
		printf (" %f %f %f %f /Helvetica 15 ",
			layout.chan_label_xmin, sip -> bot_y,
			layout.chan_label_xmax, sip -> top_y);
		emit_postscript_string (sip -> title);
		printf (" ShowTextInBox\n");
	}

	/* Draw horizontal tick lines (very very thin and light) */
	/* for the +inf, zero, -inf, and nan lines.		*/
	top_y = sip -> top_y;
	bot_y = sip -> bot_y;

	top_band = sip -> top_band;
	bot_band = sip -> bot_band;

	if ((top_band <= 0) AND (0 <= bot_band)) {
		alpha = sip -> pinf_pos;
		if (alpha < 0.0) abort ();
		y = alpha * top_y + (1.0 - alpha) * bot_y;
		/* Note: this displays the left and right tick marks too! */
		printf ("%% +infinity line\n"
			"gsave\n"
			" 0 setgray\n"
			" 72.0 600.0 div setlinewidth\n"
			" %f %f moveto\n"
			" %f %f lineto\n"
			" stroke\n"
			"grestore\n",
			layout.left_tick_xmin, y,
			layout.right_tick_xmax, y);
		alpha = sip -> pinf_pos - 0.03125;

		/* Draw /infinity label on this line. */
		alpha = sip -> pinf_label_top;
		y1 = alpha * top_y + (1.0 - alpha) * bot_y;
		alpha = sip -> pinf_label_bot;
		y2 = alpha * top_y + (1.0 - alpha) * bot_y;
		printf (" %f %f %f %f /ScaleFont 12 { (@) SimpleText } ShowRightTextInBox\n ",
			layout.left_scale_xmin, y2,
			layout.left_scale_xmax, y1);

		printf (" %f %f %f %f /ScaleFont 12 { (@) SimpleText } ShowLeftTextInBox\n ",
			layout.right_scale_xmin, y2,
			layout.right_scale_xmax, y1);
	}

	if ((top_band <= 1) AND (1 <= bot_band)) {
		alpha = sip -> plus_top_pos;
		if (alpha < 0.0) abort ();
		y1 = alpha * top_y + (1.0 - alpha) * bot_y;
		alpha = sip -> plus_bot_pos;
		if (alpha < 0.0) abort ();
		y2 = alpha * top_y + (1.0 - alpha) * bot_y;
		/* This does *NOT* display any tick marks. */
		printf ("gsave\n"
			" 0 setgray\n"
			" 72.0 600.0 div setlinewidth\n"
			" %f %f moveto\n"
			" %f %f lineto\n"
			" %f %f lineto\n"
			" %f %f lineto\n"
			" closepath\n"
			" stroke\n"
			"grestore\n",
			layout.plot_area_xmin, y2,
			layout.plot_area_xmax, y2,
			layout.plot_area_xmax, y1,
			layout.plot_area_xmin, y1);
		alpha = sip -> plus_label_top;
		y3 = alpha * top_y + (1.0 - alpha) * bot_y;
		alpha = sip -> plus_label_bot;
		y4 = alpha * top_y + (1.0 - alpha) * bot_y;
		plot_y_scale (sip, &(sip -> pband), 1, y1, y2, y3, y4);
	}

	if ((top_band <= 2) AND (2 <= bot_band)) {
		alpha = sip -> zero_pos;
		if (alpha < 0.0) abort ();
		y = alpha * top_y + (1.0 - alpha) * bot_y;
		/* Note: this displays the left and right tick marks too! */
		printf ("gsave\n"
			" 0 setgray\n"
			" 72.0 600.0 div setlinewidth\n"
			" %f %f moveto\n"
			" %f %f lineto\n"
			" stroke\n"
			"grestore\n",
			layout.left_tick_xmin, y,
			layout.right_tick_xmax, y);

		/* Label the zero line. */
		alpha = sip -> zero_label_top;
		y1 = alpha * top_y + (1.0 - alpha) * bot_y;
		alpha = sip -> zero_label_bot;
		y2 = alpha * top_y + (1.0 - alpha) * bot_y;
		printf (" %f %f %f %f /ScaleFont 12 { (0) SimpleText } ShowRightTextInBox\n ",
			layout.left_scale_xmin, y2,
			layout.left_scale_xmax, y1);

		printf (" %f %f %f %f /ScaleFont 12 { (0) SimpleText } ShowLeftTextInBox\n ",
			layout.right_scale_xmin, y2,
			layout.right_scale_xmax, y1);
	}

	if ((top_band <= 3) AND (3 <= bot_band)) {
		alpha = sip -> minus_top_pos;
		if (alpha < 0.0) abort ();
		y1 = alpha * top_y + (1.0 - alpha) * bot_y;
		alpha = sip -> minus_bot_pos;
		if (alpha < 0.0) abort ();
		y2 = alpha * top_y + (1.0 - alpha) * bot_y;
		/* This does *NOT* display any tick marks. */
		printf ("gsave\n"
			" 0 setgray\n"
			" 72.0 600.0 div setlinewidth\n"
			" %f %f moveto\n"
			" %f %f lineto\n"
			" %f %f lineto\n"
			" %f %f lineto\n"
			" closepath\n"
			" stroke\n"
			"grestore\n",
			layout.plot_area_xmin, y2,
			layout.plot_area_xmax, y2,
			layout.plot_area_xmax, y1,
			layout.plot_area_xmin, y1);
		alpha = sip -> minus_label_top;
		y3 = alpha * top_y + (1.0 - alpha) * bot_y;
		alpha = sip -> minus_label_bot;
		y4 = alpha * top_y + (1.0 - alpha) * bot_y;
		plot_y_scale (sip, &(sip -> mband), -1, y1, y2, y3, y4);
	}

	if ((top_band <= 4) AND (4 <= bot_band)) {
		alpha = sip -> minf_pos;
		if (alpha < 0.0) abort ();
		y = alpha * top_y + (1.0 - alpha) * bot_y;
		/* Note: this displays the left and right tick marks too! */
		printf ("gsave\n"
			" 0 setgray\n"
			" 72.0 600.0 div setlinewidth\n"
			" %f %f moveto\n"
			" %f %f lineto\n"
			" stroke\n"
			"grestore\n",
			layout.left_tick_xmin, y,
			layout.right_tick_xmax, y);

		/* Draw -/infinity label on this line. */
		alpha = sip -> minf_label_top;
		y1 = alpha * top_y + (1.0 - alpha) * bot_y;
		alpha = sip -> minf_label_bot;
		y2 = alpha * top_y + (1.0 - alpha) * bot_y;
		printf (" %f %f %f %f /ScaleFont 12 { (-@) SimpleText } ShowRightTextInBox\n ",
			layout.left_scale_xmin, y2,
			layout.left_scale_xmax, y1);

		printf (" %f %f %f %f /ScaleFont 12 { (-@) SimpleText } ShowLeftTextInBox\n ",
			layout.right_scale_xmin, y2,
			layout.right_scale_xmax, y1);
	}

	if ((top_band <= 5) AND (5 <= bot_band)) {
		alpha = sip -> nan_pos;
		if (alpha < 0.0) abort ();
		y = alpha * top_y + (1.0 - alpha) * bot_y;
		/* Note: this displays the left and right tick marks too! */
		printf ("gsave\n"
			" 0 setgray\n"
			" 72.0 600.0 div setlinewidth\n"
			" %f %f moveto\n"
			" %f %f lineto\n"
			" stroke\n"
			"grestore\n",
			layout.left_tick_xmin, y,
			layout.right_tick_xmax, y);

		/* Label the "NaN" line. */
		alpha = sip -> nan_label_top;
		y1 = alpha * top_y + (1.0 - alpha) * bot_y;
		alpha = sip -> nan_label_bot;
		y2 = alpha * top_y + (1.0 - alpha) * bot_y;
		printf (" %f %f %f %f /ScaleFont 12 { (NaN) SimpleText } ShowRightTextInBox\n ",
			layout.left_scale_xmin, y2,
			layout.left_scale_xmax, y1);
		printf (" %f %f %f %f /ScaleFont 12 { (NaN) SimpleText } ShowLeftTextInBox\n ",
			layout.left_scale_xmin, y2,
			layout.left_scale_xmax, y1);
	}


	j    = sip -> field;
	data = field_data [j];
	if (sip -> scatter) {
		/* Doing "scatter plot".  Just plot tiny little points. */
		width = (layout.data_xmax - layout.data_xmin) / denom;
		printf (" gsave\n"
			" 0 setgray\n"
			" 10 dict begin\n"
			" /H 72.0 600.0 div def\n"
			" /W %f def\n"
			" /H2 H -0.5 mul def\n"
			" /W2 W -0.5 mul def\n"
			" /DOT { %% x y\n"
			"  moveto\n"
			"  W2 H2 rmoveto\n"
			"  W 0 rlineto\n"
			"  0 H rlineto\n"
			"  W neg 0 rlineto\n"
			"  closepath\n"
			"  fill\n"
			" } def\n",
			width);
		for (i = xmin; i <= xmax; i++) {
			y = datum_y_value (sip, &data [i]);
			alpha = ((double) (i - scale_xmin)) / denom;
			x = (1.0 - alpha) * layout.data_xmin
				  + alpha * layout.data_xmax;
			printf (" %f %f DOT\n", x, y);
		}
		printf (" end\n"
			" grestore\n");
	}
	else {
		/* Normal data plot -- connect points with lines. */
		printf (" gsave\n"
			" 0 setgray\n"
			" 72.0 300.0 div setlinewidth\n");
		op = "moveto";
		denom = scale_xmax - scale_xmin;
		nseg = 0;
		for (i = xmin; i <= xmax; i++) {
			y = datum_y_value (sip, &data [i]);
			alpha = ((double) (i - scale_xmin)) / denom;
			x = (1.0 - alpha) * layout.data_xmin
				  + alpha * layout.data_xmax;
			printf (" %f %f %s\n", x, y, op);
			op = "lineto";
			++nseg;
			if (nseg >= 500) {
				/* Limit size of path: stroke and */
				/* start over. */
				nseg = 0;
				printf (" stroke\n");
				op = "moveto";
			}
		}
		if (nseg > 0) {
			printf (" stroke\n");
		}
		printf (" grestore\n");
	}
}

/*
 * Plot the scale (numbers and ticks) for the given "band" (positive or
 * negative) of the given strip.
 */

	static
	void
plot_y_scale (

struct StripInfo *	sip,		/* IN - strip we are plotting */
struct BScale *		bsp,		/* IN - band scaling info */
int			sign,		/* IN - +1 or -1 (which band) */
double			ymax,		/* IN - top of band data */
double			ymin,		/* IN - bottom of band data */
double			label_ymax,	/* IN - top of labels */
double			label_ymin	/* IN - bottom of labels */
)
{
	if (bsp -> logscale) {
		plot_log_y_scale (sip, bsp, sign, ymax, ymin, label_ymax, label_ymin);
	}
	else {
		plot_linear_y_scale (sip, bsp, sign, ymax, ymin, label_ymax, label_ymin);
	}
}

/*
 * Plot the logarithmic scale (numbers and ticks) for the given "band"
 * (positive or negative) of the given strip.
 */

	static
	void
plot_log_y_scale (

struct StripInfo *	sip,		/* IN - strip we are plotting */
struct BScale *		bsp,		/* IN - band scaling info */
int			sign,		/* IN - +1 or -1 (which band) */
double			ymax,		/* IN - top of band data */
double			ymin,		/* IN - bottom of band data */
double			label_ymax,	/* IN - top of labels */
double			label_ymin	/* IN - bottom of labels */
)
{
int		i;
const char *	sign_str;
double		y, y1, y2, dy, dy2;
double		alpha, denom, temp;

	dy = label_ymax - ymax;

	sign_str = "";
	if (sign < 0) {
		/* Negative band -- plot "upside down". */
		temp = ymax;
		ymax = ymin;
		ymin = temp;
		sign_str = "-";
	}

	printf (" %% Draw Y scale for %s band\n",
		sign < 0 ? "negative" : "positive");

	printf (" 0 setgray\n"
		" 72.0 600.0 div setlinewidth\n");

	denom = bsp -> maxpow - bsp -> minpow;

	/* The size of the label boxes need to be a compromise between	*/
	/* the +dy and -dy values above (which restrict how far the	*/
	/* top-most and bottom-most labels can protrude beyond the data	*/
	/* area and into the space between strips or the bands/lines of	*/
	/* single data channel's strip), and the space between labels	*/
	/* in the middle.  For example, it is often the case that the	*/
	/* labels inside the band are widely separated (thus allowing	*/
	/* for large label text), yet tiny spacing between the bands/	*/
	/* lines cause a very small dy value above -- and very tiny	*/
	/* text labels even when there is space for big labels.		*/
	/*								*/
	/* To address this problem, we now compute dy2, which is the	*/
	/* distance between labels inside the scale.  If it turns out	*/
	/* that dy < dy2, then instead of making text boxes that are	*/
	/* 2*dy units tall, we make boxes that are dy+dy2 units tall,	*/
	/* and simply let the top-most and bottom-most boxes slide	*/
	/* inward, thereby being no longer centered on the tick mark	*/
	/* they are labeling.  This will permit the text to be larger	*/
	/* and make better use of the interior space available.		*/

	alpha = bsp -> powtics / denom;
	dy2 = 0.45 * alpha * fabs (ymax - ymin);
	if (dy < dy2) {
		/* Make the text boxes a bit taller than they would	*/
		/* otherwise be.  The outermost ones will have to	*/
		/* slide inwards a bit to compensate...			*/
		dy = 0.5 * (dy + dy2);
	}


	for (i = bsp -> minpow; i <= bsp -> maxpow; i += bsp -> powtics) {
		/* Draw left tick. */
		alpha = (i - bsp -> minpow) / denom;
		y = (1.0 - alpha) * ymin + alpha * ymax;
		printf (" %f %f moveto\n"
			" %f %f lineto\n",
			layout.left_tick_xmin, y,
			layout.left_tick_xmax, y);

		/* Draw left label. */
		y1 = y + dy;
		y2 = y - dy;
		if (y1 > label_ymax) {
			/* Slide box downward. */
			y1 = label_ymax;
			y2 = y1 - 2.0 * dy;
		}
		if (y2 < label_ymin) {
			/* Slide box upward. */
			y2 = label_ymin;
			y1 = y2 + 2.0 * dy;
		}
		printf (" %f %f %f %f /ScaleFont 12\n"
			"  { (%s10) (%d) SuperscriptText } ShowRightTextInBox\n",
			layout.left_scale_xmin, y2,
			layout.left_scale_xmax, y1,
			sign_str,
			i);

		/* Draw right tick. */
		printf (" %f %f moveto\n"
			" %f %f lineto\n",
			layout.right_tick_xmin, y,
			layout.right_tick_xmax, y);

		/* Draw right label. */
		printf (" %f %f %f %f /ScaleFont 12\n"
			"  { (%s10) (%d) SuperscriptText } ShowLeftTextInBox\n",
			layout.right_scale_xmin, y2,
			layout.right_scale_xmax, y1,
			sign_str,
			i);
	}
}

/*
 * Plot the linear scale (numbers and ticks) for the given "band"
 * (positive or negative) of the given strip.
 */

	static
	void
plot_linear_y_scale (

struct StripInfo *	sip,		/* IN - strip we are plotting */
struct BScale *		bsp,		/* IN - band scaling info */
int			sign,		/* IN - +1 or -1 (which band) */
double			ymax,		/* IN - top of band data */
double			ymin,		/* IN - bottom of band data */
double			label_ymax,	/* IN - top of labels */
double			label_ymin	/* IN - bottom of labels */
)
{
int		i;
double		yval;
double		y, y1, y2, dy, dy2;
double		alpha, denom, temp;

	dy = label_ymax - ymax;

	if (sign < 0) {
		/* Negative band -- plot "upside down". */
		temp = ymax;
		ymax = ymin;
		ymin = temp;
	}

	printf (" %% Draw Y scale for %s band\n",
		sign < 0 ? "negative" : "positive");

	printf (" 0 setgray\n"
		" 72.0 600.0 div setlinewidth\n");

	denom = bsp -> maxval - bsp -> minval;

	/* The size of the label boxes need to be a compromise between	*/
	/* the +dy and -dy values above (which restrict how far the	*/
	/* top-most and bottom-most labels can protrude beyond the data	*/
	/* area and into the space between strips or the bands/lines of	*/
	/* single data channel's strip), and the space between labels	*/
	/* in the middle.  For example, it is often the case that the	*/
	/* labels inside the band are widely separated (thus allowing	*/
	/* for large label text), yet tiny spacing between the bands/	*/
	/* lines cause a very small dy value above -- and very tiny	*/
	/* text labels even when there is space for big labels.		*/
	/*								*/
	/* To address this problem, we now compute dy2, which is the	*/
	/* distance between labels inside the scale.  If it turns out	*/
	/* that dy < dy2, then instead of making text boxes that are	*/
	/* 2*dy units tall, we make boxes that are dy+dy2 units tall,	*/
	/* and simply let the top-most and bottom-most boxes slide	*/
	/* inward, thereby being no longer centered on the tick mark	*/
	/* they are labeling.  This will permit the text to be larger	*/
	/* and make better use of the interior space available.		*/

	alpha = bsp -> ytics / denom;
	dy2 = 0.45 * alpha * fabs (ymax - ymin);
	if (dy < dy2) {
		/* Make the text boxes a bit taller than they would	*/
		/* otherwise be.  The outermost ones will have to	*/
		/* slide inwards a bit to compensate...			*/
		dy = 0.5 * (dy + dy2);
	}


	for (yval = bsp -> minval; yval <= bsp -> maxval; yval += bsp -> ytics) {
		/* Draw left tick. */
		alpha = (yval - bsp -> minval) / denom;
		y = (1.0 - alpha) * ymin + alpha * ymax;
		printf (" %f %f moveto\n"
			" %f %f lineto\n",
			layout.left_tick_xmin, y,
			layout.left_tick_xmax, y);

		/* Draw left label. */
		y1 = y + dy;
		y2 = y - dy;
		if (y1 > label_ymax) {
			/* Slide box downward. */
			y1 = label_ymax;
			y2 = y1 - 2.0 * dy;
		}
		if (y2 < label_ymin) {
			/* Slide box upward. */
			y2 = label_ymin;
			y1 = y2 + 2.0 * dy;
		}
		printf (" %f %f %f %f /ScaleFont 12\n"
			"  { (%g) SimpleText } ShowRightTextInBox\n",
			layout.left_scale_xmin, y2,
			layout.left_scale_xmax, y1,
			sign * yval);

		/* Draw right tick. */
		printf (" %f %f moveto\n"
			" %f %f lineto\n",
			layout.right_tick_xmin, y,
			layout.right_tick_xmax, y);

		/* Draw right label. */
		printf (" %f %f %f %f /ScaleFont 12\n"
			"  { (%g) SimpleText } ShowLeftTextInBox\n",
			layout.right_scale_xmin, y2,
			layout.right_scale_xmax, y1,
			sign * yval);
	}
}

/*
 * Determine where on the page the given Y coordinate belongs for the
 * given strip.
 */

	static
	double
datum_y_value (

struct StripInfo *	sip,	/* IN - strip we are plotting */
const struct Datum *	dp	/* IN - datum to translate into Y coord */
)
{
int		type;
double		alpha, beta, val;
double		y;
double		top_y, bot_y;

	top_y = sip -> top_y;
	bot_y = sip -> bot_y;

	type = dp -> type;

	if ((sip -> force_linear) AND (type EQ DATUM_TYPE_VALUE)) {
		/* Forced linear scale.  Convert out-of-range	*/
		/* values to +inf or -inf.			*/
		val = dp -> value;
		if (val > sip -> linear_max) {
			type = DATUM_TYPE_PINF;
		}
		else if (val < sip -> linear_min) {
			type = DATUM_TYPE_MINF;
		}
	}

	switch (type) {
	case DATUM_TYPE_VALUE:
		break;

	case DATUM_TYPE_PINF:
		alpha = sip -> pinf_pos;
		if (alpha < 0.0) abort ();
		y = alpha * top_y + (1.0 - alpha) * bot_y;
		return (y);

	case DATUM_TYPE_MINF:
		alpha = sip -> minf_pos;
		if (alpha < 0.0) abort ();
		y = alpha * top_y + (1.0 - alpha) * bot_y;
		return (y);

	case DATUM_TYPE_NAN:
		alpha = sip -> nan_pos;
		if (alpha < 0.0) abort ();
		y = alpha * top_y + (1.0 - alpha) * bot_y;
		return (y);

	default:
		abort ();
	}

	/* We have a numeric value */
	val = dp -> value;
	if (val EQ 0.0) {
		alpha = sip -> zero_pos;
		if (alpha >= 0.0) {
			y = alpha * top_y + (1.0 - alpha) * bot_y;
			return (y);
		}
		if ((sip -> pband.logscale EQ 0) AND
		    (sip -> pband.minval <= 0.0)) {
			beta = scale_numeric_value (val, &(sip -> pband));
			alpha = (1.0 - beta) * sip -> plus_bot_pos + beta * sip -> plus_top_pos;
			y = alpha * top_y + (1.0 - alpha) * bot_y;
			return (y);
		}
		else {
			abort ();
		}
	}

	if (val > 0) {
		/* Get [0,1] value where 0 ==> pmin, 1 ==> pmax. */
		beta = scale_numeric_value (val, &(sip -> pband));
		alpha = (1.0 - beta) * sip -> plus_bot_pos + beta * sip -> plus_top_pos;
		y = alpha * top_y + (1.0 - alpha) * bot_y;
		return (y);
	}

	/* Negative value. */
	/* Get [0,1] value where 0 ==> mmin, 1 ==> mmax. */
	beta = scale_numeric_value (-val, &(sip -> mband));
	alpha = (1.0 - beta) * sip -> minus_top_pos + beta * sip -> minus_bot_pos;
	y = alpha * top_y + (1.0 - alpha) * bot_y;
	return (y);
}

/*
 * Scale a value according to the settings of the given numeric "band".
 */

	static
	double
scale_numeric_value (

double		val,		/* IN - positive value to scale */
struct BScale *	bsp		/* IN - band scaling info */
)
{
int		n;
double		alpha, lmag;

	if (NOT (bsp -> logscale)) {
		/* We have a linear scale. */
		alpha = (val - bsp -> minval) /
			(bsp -> maxval - bsp -> minval);
		if ((alpha < 0.0) OR (alpha > 1.0)) abort ();
		return (alpha);
	}

	/* Log scaled */
	n = bsp -> maxpow - bsp -> minpow;

	lmag = log10 (val);
	lmag -= bsp -> minpow;

	alpha = lmag / n;

	if ((alpha < 0.0) OR (alpha > 1.0)) abort ();

	return (alpha);
}

/*
 * Draw tick marks along the X axis along the top of the top strip,
 * and along the bottom of the bottom strip.  Also draw numeric labels
 * for these ticks along the bottom.
 */

	static
	void
plot_x_ticks_and_scale (void)

{
struct StripInfo *	sip;
double			xval;
double			x, dx;
double			alpha, denom;
double			top_strip_ymax;
double			bot_strip_ymin;

	printf (" %% Draw X-axis ticks and scale numbering.\n");
	printf (" gsave\n"
		"  72.0 300.0 div setlinewidth\n");

	sip = &strip_info [0];
	top_strip_ymax = sip -> top_y;

	sip = &strip_info [nstrips - 1];
	bot_strip_ymin = sip -> bot_y;

	denom = scale_xmax - scale_xmin;

	alpha = 0.45 * scale_xtics / denom;
	dx = alpha * (layout.data_xmax - layout.data_xmin);

	for (xval = scale_xmin; xval <= scale_xmax; xval += scale_xtics) {
		/* Draw top tick. */
		alpha = (xval - scale_xmin) / denom;
		x = (1.0 - alpha) * layout.data_xmin + alpha * layout.data_xmax;
		printf ("  %f %f moveto\n"
			"  0 4.5 rlineto\n"
			"  stroke\n",
			x, top_strip_ymax);

		/* Draw bottom tick. */
		printf ("  %f %f moveto\n"
			"  0 -4.5 rlineto\n"
			"  stroke\n",
			x, bot_strip_ymin);

		/* Draw label below bottom tick mark. */
		printf ("  %f %f %f %f /ScaleFont 12 "
			"{ (%g) SimpleText } ShowCenteredTextInBox\n",
			x - dx, layout.bottom_scale_ymin,
			x + dx, layout.bottom_scale_ymax,
			xval);
	}
	printf (" grestore\n");
}

/*
 * Emit a postscript string constant.  We must be careful to escape
 * certain characters.
 */

	static
	void
emit_postscript_string (

const char *	str
)
{
int		c;
int		must_escape;

	printf ("(");
	for (;;) {
		c = *str++ & 0xFF;
		if (c EQ '\0') break;
		if ((c < ' ') OR ('~' < c) OR
		    (c EQ '(') OR (c EQ ')') OR
		    (c EQ '\\')) {
			printf ("\\%03o", c);
		}
		else {
			putchar (c);
		}
	}
	printf (")");
}

/*
 * Draw a light gray box.
 */

	static
	void
box (

double		xmin,
double		xmax,
double		ymin,
double		ymax
)
{
	xmin += 72.0;
	printf (" %f %f moveto\n", xmin, ymin);
	printf (" %f %f lineto\n", xmax, ymin);
	printf (" %f %f lineto\n", xmax, ymax);
	printf (" %f %f lineto\n", xmin, ymax);
	printf (" closepath\n");
	printf (" 0.95 setgray\n");
	printf (" fill\n");
}
