%{
/***********************************************************************

	File:	lex.l
	Rev:	a-2
	Date:	09/02/2009

	Copyright (c) 2006, 2009 by David M. Warme.  All rights
	reserved.

************************************************************************

	Simple flex lexer for eeg configuration file.

************************************************************************

	Modification Log:

	a-1:	05/29/2006	warme
		: Created.
	a-2:	09/02/2009	warme
		: Added new tokens for "linear" channel option.

************************************************************************/


%}
%Start A com

%{
#include "config.tab.h"
#include "logic.h"
#include "name.h"
#include <string.h>

int	lineno;

#define YY_SKIP_YYWRAP
#define yywrap()	1	/* No file-wrapping: EOF is FINAL! */
#define	ECHO		{ }	/* A null statement: never echo anything! */
#define YY_NEVER_INTERACTIVE	1
#define	YY_USER_INIT	init_keywords ()

struct keyword {
	const char *	text;
	int		token;
};

const struct keyword	keywords [] = {
	"bottom_title",		BOTTOM_TITLE,
	"chan",			CHAN,
	"data",			DATA,
	"id",			ID,
	"landscape",		LANDSCAPE,
	"linear",		LINEAR,
	"mag",			MAGNIFY,
	"magnify",		MAGNIFY,
	"orientation",		ORIENTATION,
	"plot",			PLOT,
	"portrait",		PORTRAIT,
	"scatter",		SCATTER,
	"top_title",		TOP_TITLE,
	"version",		VERSION,
	NULL,			0,
};

	static
	void
init_keywords (void)

{
const struct keyword *	kp;
const char *		cp;
struct Name *		np;

	kp = &keywords [0];
	for ( ; ; kp++) {
		cp = kp -> text;
		if (cp EQ NULL) break;
		np = make_name (cp);
		np -> token = kp -> token;
	}
}

	static
	int
identify_keyword (

const char *	p,
int		n,
YYSTYPE *	lvalp
)
{
int		i;
const char *	endp;
struct Name *	np;

	endp = p + n;
	np = find_name_lower_case (p, endp);
	i = 0;
	if (np NE NULL) {
		i = np -> token;
	}

	return (i);
}

	static
	struct Name *
scan_string (

char *		p,		/* IN - start of string text */
char *		endp		/* IN - end of string text */
)
{
char *		src;
char *		dst;
struct Name *	np;
char		c;

	dst = p;
	for (src = p; src < endp; ) {
		c = *src++;
		if (c EQ '\\') {
			c = *src++;
		}
		*dst++ = c;
	}
	np = make_name_from_buffer (p, dst);

	return (np);
}

%}

WS	[ \t\f\013]
DEC	[0123456789]
ID1	[ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_]
IDN	[0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_]

EXP	([eE]([+-])?{DEC}+)

%%

	/* A simple way to zero the entire yylval object... */
	yylval.y_double = 0.0;
	BEGIN A;

<A>\n		{ ++lineno; }

<A>{WS}+	;
<A>"/*"		{ BEGIN com; }

<com>\n		{ ++lineno; }
<com>"*/"	{ BEGIN A; }
<com>[^\n]	;

<A>"//"[^\n]*\n	{ ++lineno; }

<A>"+"		{ return (PLUS); }
<A>"-"		{ return (MINUS); }
<A>","		{ return (COMMA); }
<A>":"		{ return (COLON); }
<A>"("		{ return (LPAREN); }
<A>")"		{ return (RPAREN); }
<A>"{"		{ return (LBRACE); }
<A>"}"		{ return (RBRACE); }
<A>"["		{ return (LBRACK); }
<A>"]"		{ return (RBRACK); }

<A>{ID1}{IDN}*	{ int i;
		  struct Name * p;
		  yylval.y_long = 0;
		  i = identify_keyword (yytext, yyleng, &yylval);
		  if (i EQ 0) {
			/* Not a keyword -- it's an identifier. */
			i = IDENT;
			/* FIXME - make a Name. */
			p = NULL;
			yylval.y_name = p;
		  }
		  return (i);
		}

<A>{DEC}+	{ long n;
		  int i;
		  n = 0;
		  for (i = 0; i < yyleng; i++) {
			n = 10 * n + (yytext [i] - '0');
		  }
		  yylval.y_long = n;
		  return (ICONST);
		}

<A>((({DEC}*"."{DEC}+)|({DEC}+"."))({EXP})?)|({DEC}+{EXP})	{
		  char * buf;
		  char * end;
		  buf = malloc (yyleng + 1);
		  memcpy (buf, yytext, yyleng);
		  buf [yyleng] = '\0';
		  yylval.y_double = strtod (buf, &end);
		  if (end NE &buf [yyleng]) abort ();
		  free (buf);
		  return (FCONST);
		}

<A>\"([^\"\\\n]|(\\[^\n]))*\"		{
		  yylval.y_name = scan_string (yytext + 1, yytext + (yyleng-1));

		  return (STRING);
		}
<A>\"([^\"\\\n]|(\\[^\n]))*\n	{
		  fprintf (stderr, "Newline in string at line %d\n", lineno);

		  yylval.y_name = scan_string (yytext + 1, yytext + (yyleng-1));
		  return (STRING);
		}

<A>.		{ fprintf (stderr, "Illegal character on line %d\n", lineno);
		}
