/***********************************************************************

	$Id: name.c,v 1.5 2009/09/02 10:31:08 warme Exp $

	File:	name.c
	Rev:	a-1
	Date:	05/29/2006

	Copyright (c) 2006, 2009 by David M. Warme.  All rights
	reserved.

************************************************************************

	A simple hashed symbol table of names.

************************************************************************

	Modification Log:

	a-1:	05/29/2006	warme
		: Created.

************************************************************************/

#include "name.h"

#include "logic.h"
#include "memory.h"
#include <string.h>

/*
 * Global Routines
 */

struct Name *	find_name_lower_case (const char * p, const char * end);
struct Name *	make_name (const char * p);
struct Name *	make_name_from_buffer (const char * p, const char * end);


/*
 * Local Stuff
 */

#define	HASH_TABLE_SIZE		1009	/* Yes, it's a prime number. */

typedef   signed long	int32s;
typedef unsigned long	int32u;

static int	compare_lower_case (const char * p1, const char * p2, size_t n);
static int32u	hash_buffer (const char * p, const char * end);
static int32u	hash_buffer_lower_case (const char * p, const char * end);
static int32u	hash_string (const char * p, size_t * length);

static struct Name *	hash_table [HASH_TABLE_SIZE];

/*
 * Return the Name whose text matches the given null-terminated string.
 * If no such Name exists yet, make a new one, add it to the hash table,
 * and return it.
 */

	struct Name *
make_name (

const char *	p	/* IN - null-terminated string to make name for */
)
{
int32u		h;
size_t		len;
struct Name **	hookp;
struct Name *	np;
char *		cp;

	h = hash_string (p, &len);

	hookp = &hash_table [h];
	for (np = *hookp; np NE NULL; np = np -> next) {
		if ((len EQ np -> len) AND
		    (memcmp (np -> text, p, len) EQ 0)) {
			return (np);
		}
	}

	np = NEW (struct Name);
	cp = NEWA (len + 1, char);
	memcpy (cp, p, len);
	cp [len] = '\0';

	np -> text	= cp;
	np -> len	= len;
	np -> hash	= h;
	np -> token	= 0;
	np -> fldno	= -1;
	np -> next	= *hookp;

	*hookp = np;

	return (np);
}

/*
 * Return the Name whose text matches the given buffer.
 * If no such Name exists yet, make a new one, add it to the hash table,
 * and return it.
 */

	struct Name *
make_name_from_buffer (

const char *	p,	/* IN - start of string to make name for */
const char *	end	/* IN - end of string to make name for */
)
{
int32u		h;
size_t		len;
struct Name **	hookp;
struct Name *	np;
char *		cp;

	h = hash_buffer (p, end);

	len = end - p;

	hookp = &hash_table [h];
	for (np = *hookp; np NE NULL; np = np -> next) {
		if ((len EQ np -> len) AND
		    (memcmp (np -> text, p, len) EQ 0)) {
			return (np);
		}
	}

	np = NEW (struct Name);
	cp = NEWA (len + 1, char);
	memcpy (cp, p, len);
	cp [len] = '\0';

	np -> text	= cp;
	np -> len	= len;
	np -> hash	= h;
	np -> token	= 0;
	np -> fldno	= -1;
	np -> next	= *hookp;

	*hookp = np;

	return (np);
}

/*
 * Return the Name whose text matches the "forced-to-be-lower-case-version"
 * of the text in the given buffer.  Return NULL if no such Name currently
 * exists.
 */

	struct Name *
find_name_lower_case (

const char *	p,	/* IN - start of string to make name for */
const char *	end	/* IN - end of string to make name for */
)
{
int32u		h;
size_t		len;
struct Name **	hookp;
struct Name *	np;

	h = hash_buffer_lower_case (p, end);

	len = end - p;

	hookp = &hash_table [h];
	for (np = *hookp; np NE NULL; np = np -> next) {
		if ((len EQ np -> len) AND
		    (compare_lower_case (p, np -> text, len) EQ 0)) {
			return (np);
		}
	}
	return (NULL);
}

/*
 * Compare two fixed-length byte sequences, the first of which we force
 * to lower-case before comparing.
 */

	static
	int
compare_lower_case (

const char *	p1,		/* IN - first string (force to lower case) */
const char *	p2,		/* IN - second string */
size_t		n		/* IN - number of chars to compare */
)
{
unsigned char	c1;
unsigned char	c2;

	for ( ; n > 0; n--) {
		c1 = *p1++;
		c2 = *p2++;
		/* FIXME - this assumes something like ASCII. */
		if ((c1 >= 'A') AND (c1 <= 'Z')) {
			c1 += ('a' - 'A');
		}
		if (c1 NE c2) return (c1 - c2);
	}
	return (0);
}

/*
 * Hash the given null-terminated string, and also return its length.
 */

	static
	int32u
hash_string (

const char *		string,		/* IN - string to hash */
size_t *		length		/* OUT - length of string */
)
{
size_t			len;
int			c1;
int			c2;
int32u			reg;
const unsigned char *	cp;

	/* We want UNSIGNED characters for this... */
	cp = (const unsigned char *) string;

	reg = 0;
	len = 0;
	for (;;) {
		c1 = *cp++;
		if (c1 EQ '\0') break;

		/* Rotate reg left 1 bit... */
		if (((int32s) reg) < 0) {
			reg = (reg << 1) + 1;
		}
		else {
			reg = (reg << 1);
		}
		c2 = *cp++;
		reg ^= (c1 << 8) | c2;
		if (c2 EQ '\0') {
			++len;
			break;
		}
		len += 2;
	}

	*length = len;

	reg %= ((int32u) HASH_TABLE_SIZE);

	return (reg);
}

/*
 * Hash the text in the given buffer.
 */

	static
	int32u
hash_buffer (

const char *		begin,		/* IN - start of text to hash */
const char *		end		/* IN - end of text to hash */
)
{
int			c1;
int			c2;
int32u			reg;
const unsigned char *	cp;
const unsigned char *	endp;

	/* We want UNSIGNED characters for this... */
	cp   = (const unsigned char *) begin;
	endp = (const unsigned char *) end;

	reg = 0;
	for (;;) {
		if (cp >= endp) break;
		c1 = *cp++;

		/* Rotate reg left 1 bit... */
		if (((int32s) reg) < 0) {
			reg = (reg << 1) + 1;
		}
		else {
			reg = (reg << 1);
		}
		if (cp >= endp) {
			reg ^= (c1 << 8);
			break;
		}
		c2 = *cp++;
		reg ^= ((c1 << 8) | c2);
	}

	reg %= ((int32u) HASH_TABLE_SIZE);

	return (reg);
}

/*
 * Hash the "lower-case version" of the text in the given buffer.
 */

	static
	int32u
hash_buffer_lower_case (

const char *		begin,		/* IN - start of text to hash */
const char *		end		/* IN - end of text to hash */
)
{
int			c1;
int			c2;
int32u			reg;
const unsigned char *	cp;
const unsigned char *	endp;

	/* We want UNSIGNED characters for this... */
	cp   = (const unsigned char *) begin;
	endp = (const unsigned char *) end;

	reg = 0;
	for (;;) {
		if (cp >= endp) break;
		c1 = *cp++;
		/* FIXME - this assumes something like ASCII... */
		if ((c1 >= 'A') AND (c1 <= 'Z')) {
			c1 += ('a' - 'A');
		}

		/* Rotate reg left 1 bit... */
		if (((int32s) reg) < 0) {
			reg = (reg << 1) + 1;
		}
		else {
			reg = (reg << 1);
		}
		if (cp >= endp) {
			reg ^= (c1 << 8);
			break;
		}
		c2 = *cp++;
		/* FIXME - this assumes something like ASCII... */
		if ((c2 >= 'A') AND (c2 <= 'Z')) {
			c2 += ('a' - 'A');
		}
		reg ^= ((c1 << 8) | c2);
	}

	reg %= ((int32u) HASH_TABLE_SIZE);

	return (reg);
}
